"""
Quantiphyse - GUI widgets for easily defining options

The intention is that this module will supercede most of the contents
of quantiphyse.gui.widgets which mostly contains thin wrappers around
Qt widgets.

The idea is that you create an OptionBox and then use ``add()`` to
add any of the option widgets to it, providing a label and generally 
a key. The ``values()`` method then returns a dictionary of option
values suitable to feeding in to a ``Process``. 

Alternatively options can be used directly. All have a ``value``
property which returns an appropriate value, and a ``sig_changed`` signal
which is emitted whenever this value changes

Copyright (c) 2013-2018 University of Oxford
"""
import os
import logging
import collections

import six
import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.utils import QpException, sf, load_matrix, local_file_from_drop_url, default_save_dir, set_default_save_dir
from quantiphyse.gui.dialogs import MatrixViewerDialog
from quantiphyse.gui.pickers import PickMode

LOG = logging.getLogger(__name__)

class OptionBox(QtGui.QGroupBox):
    """
    A box containing structured options for a QpWidget
    """
    sig_changed = QtCore.Signal()

    def __init__(self, title=""):
        QtGui.QGroupBox.__init__(self, title)
        if not title:
            self.setStyleSheet("border: none")
        self.grid = QtGui.QGridLayout()
        self.setLayout(self.grid)
        self._current_row = 0
        self._options = {}
        self._rows = {}

    def add(self, label, *options, **kwargs):
        """
        Add labelled option widgets to the option box

        :param label: Text label for the option widgets
        :param options: Sequence of arguments, each a QtWidget
        :param key: Optional keyword argument specifying the key to use when 
                    returning the option value in ``values()``. Only valid
                    when a single option is given. If not specified, the 
                    label is used.
        :param keys: Optional keyword argument specifying a sequence of keys
                     to use when returning the option values in ``values()``.
                     Must be specified when there are multiple options, and
                     must have the same length as the number of options.
        """
        checked = kwargs.get("checked", False)
        enabled = kwargs.get("enabled", False)
        key = kwargs.get("key", label)
        keys = kwargs.get("keys", [key,])
        
        real_options = [opt for opt in options if isinstance(opt, Option)]
        extra_widgets = [opt for opt in options if opt not in real_options]

        if not real_options:
            # Allow no options for just a label
            real_options = []
            keys = []

        if len(keys) != len(real_options):
            raise ValueError("keys must be sequence which is the same length as the number of options")

        self.grid.addWidget(QtGui.QLabel(label), self._current_row, 0, 1, 1 if real_options else 3)
        opt_col = 1
        if checked:
            cb = QtGui.QCheckBox()
            cb.setChecked(enabled)
            self.grid.addWidget(cb, self._current_row, 1)
            opt_col += 1

        for idx, keyopt in enumerate(zip(keys, real_options)):
            key, option = keyopt
            if option is None: continue
            LOG.debug("Adding option: %s (key=%s)", option, key)
            self.grid.addWidget(option, self._current_row, idx+opt_col, 1, 1 if checked else 2)
            if checked:
                option.setEnabled(enabled)
                cb.stateChanged.connect(self._cb_toggled(option))
            self._options[key] = option
            option.sig_changed.connect(self.sig_changed.emit)

        for extra_idx, widget in enumerate(extra_widgets):
            self.grid.addWidget(widget, self._current_row, len(real_options)+extra_idx+2)

        self._rows[key] = self._current_row
        self._current_row += 1

        if len(real_options) == 1:
            return real_options[0]
        else:
            return real_options
    
    def set_visible(self, key, visible=True):
        """
        Show or hide an option. 
        
        FIXME this only works for simple key/single option rows at the moment
        """
        row = self._rows[key]
        col = 0
        checked = True
        while 1:
            item = self.grid.itemAtPosition(row, col)
            if item is None:
                break
            item.widget().setVisible(visible)
            if col == 1 and isinstance(item.widget(), QtGui.QCheckBox) and not isinstance(item.widget(), Option):
                checked = item.widget().isChecked()
                item.widget().setEnabled(visible)
            else:
                item.widget().setEnabled(visible and checked)
            col += 1

    def set_checked(self, key, checked):
        """
        Set whether an option with a checkbox is checked or not
        """
        row = self._rows[key]
        cb = self.grid.itemAtPosition(row, 1)
        if isinstance(cb.widget(), QtGui.QCheckBox) and not isinstance(cb.widget(), Option):
            cb.widget().setChecked(checked)
        else:
            raise ValueError("set_checked called on option '%s' which is not a checked option" % key)

    def clear(self):
        """
        Clear all widgets out of the options box
        """
        while self.grid.count():
            child = self.grid.takeAt(0)
            if child.widget():
                child.widget().deleteLater()
        self._current_row = 0
        self._options = {}

    def option(self, key):
        """
        :return: The option widget having key=``key``
        """
        return self._options[key]

    def values(self):
        """
        Get the values of all options as a dict

        :return: dict of key:option value for all options
        """
        ret = collections.OrderedDict()
        for key, option in self._options.items():
            if option.isEnabled() and option.value is not None:
                ret[key] = option.value
        return ret

    def _cb_toggled(self, option):
        def _toggled(state):
            option.setEnabled(state)
            option.sig_changed.emit()
        return _toggled

class Option(object):
    """
    Base class for  an option

    We have a base class mainly so we can use ``isinstance`` to detect option
    widgets.
    
    All option widgets *must* define a ``value`` property method which returns the
    current value of the option in whatever type makes sense to the particular
    option type. Options must also define a Qt signal ``sig_changed`` which takes
    no parameters and is emitted whenever the option value changes.

    We would like to define sig_changed in the base class but this requires
    the class to be a QObject and we would end up with 
    multiple inheritance from QObject, QWidget which is probably
    very bad if it gets into the C++
    
    Options should probably make the ``value`` property settable so the option
    UI can be updated programatically.
    
    Options may support other properties, e.g. ``choices`` to set the current set
    of options in a ChoiceOption.
    """
    pass

class DataOption(Option, QtGui.QComboBox):
    """
    A combo box which gives a choice of data

    There is quite a lot of ugliness here because the box can be put into two modes - 
    single and multi-select. In multi-select mode we subvert the QComboBox to provide
    essentially a dropdown list of checkboxes. In single select mode things can be 
    complicated by providing 'all' and 'none' option.

    Probably this class should be split into two for each use case.
    """
    sig_changed = QtCore.Signal()

    def __init__(self, ivm, parent=None, **kwargs):
        super(DataOption, self).__init__(parent)
        self.ivm = ivm
        self._changed = False
        self._current_value = []
        self._all_data_names = []

        self._include_3d = kwargs.get("include_3d", True)
        self._include_4d = kwargs.get("include_4d", not kwargs.get("static_only", False))
        self._include_rois = kwargs.get("rois", False)
        self._include_nonrois = kwargs.get("data", True)
        
        self._none_option = kwargs.get("none_option", False)
        self._all_option = kwargs.get("all_option", False)
        self._multi = kwargs.get("multi", False)
        self._explicit = kwargs.get("explicit", False)
        self._data_changed()

        self.currentIndexChanged.connect(self._index_changed)
        self.ivm.sig_all_data.connect(self._data_changed)
        if self._multi:
            self.view().pressed.connect(self._item_pressed)
            delegate = QtWidgets.QStyledItemDelegate(self.view())
            self.view().setItemDelegate(delegate)

    @property
    def value(self):
        """ 
        Get the names of the selected data item(s)
        
        If neither all option nor multi options are specified, returns
        name of currently selected data item. Otherwise returns list of
        all selected data items.
        
        Also returns None if the none option is active and selected.
        """
        if self._multi:
            ret = []
            for idx in range(1, self.count()):
                item = self.model().item(idx, 0)
                if item.checkState() == QtCore.Qt.Checked:
                    ret.append(self.itemText(idx))
        else:
            current = self.currentText()
            if self._none_option and current == "<none>":
                ret = None
            elif self._explicit and current == "":
                ret = None
            elif self._all_option and current == "<all>":
                ret = list(self.ivm.data.keys())
            elif self._all_option:
                ret = [current,]
            else:
                ret = current
        return ret
        
    @value.setter
    def value(self, val):
        """ 
        Set the selected data name(s)

        In multi-select mode val should be a sequence of data item names.
        In single-select mode val should be a data item name, or None
        or '<all>' (valid if the none/all options are enabled)
        """ 
        if self._multi:
            for name in val:
                idx = self.findText(name)
                item = self.model().item(idx, 0)
                item.setCheckState(QtCore.Qt.Checked)
        else:
            if val is None and self._none_option:
                val = "<none>"
                
            if self._explicit and val is None:
                self.setCurrentIndex(-1)
            elif isinstance(val, six.string_types):
                idx = self.findText(val)
                if idx >= 0:
                    self.setCurrentIndex(idx)
                else:
                    raise ValueError("Data item %s is not a valid choice for this option")
            else:
                raise ValueError("Can't specify multiple data items when DataOption is not in multi select mode")
        self._current_value = self.value

    def _item_pressed(self, idx):
        item = self.model().itemFromIndex(idx)
        if item.checkState() == QtCore.Qt.Checked:
            item.setCheckState(QtCore.Qt.Unchecked)
        else:
            item.setCheckState(QtCore.Qt.Checked)
        self.setItemText(0, self._visible_text(self.value))
        self.sig_changed.emit()
        self._changed = True

    def hidePopup(self):
        """
        Overridden from QtGui.QComboBox

        To allow multi-select, don't hide the popup when it's clicked on to
        select/deselect data sets, so we can check and uncheck
        them in one go. However if nothing has changed as
        a result of the click (e.g. we click outside the popup
        window), this will close the popup
        """
        if not self._changed:
            QtGui.QComboBox.hidePopup(self)
        self._changed = False

    def _visible_text(self, selected_items):
        if selected_items:
            return ", ".join(selected_items)
        else:
            return "<Select data items>"

    def _index_changed(self, idx):
        if self._multi:
            self.setCurrentIndex(0)
        self._update_highlight()
        new_value = self.value
        if self._current_value != new_value:
            self._current_value = new_value
            self.sig_changed.emit()

    def _data_changed(self):
        all_data_names = list(self.ivm.data.keys())
        if all_data_names and all_data_names == self._all_data_names:
            return
        
        self._all_data_names = all_data_names
        current = self.value
        self.blockSignals(True)
        try:
            data = []
            for name, qpd in self.ivm.data.items():
                if self._include_nonrois or (qpd.roi and self._include_rois):
                    data.append(name)

            self.clear()
            
            if self._none_option and not self._multi:
                self.addItem("<none>")
            elif self._multi:
                self.addItem(self._visible_text(current))

            idx = 1
            for name in sorted(data):
                data = self.ivm.data.get(name, None)
                if data.nvols == 1 and self._include_3d:
                    self.addItem(data.name)
                    added = True
                elif data.nvols > 1 and self._include_4d:
                    self.addItem(data.name)
                    added = True
                else:
                    added = False

                if added:
                    if self._multi:
                        item = self.model().item(idx, 0)
                        if data.name in current:
                            item.setCheckState(QtCore.Qt.Checked)
                        else:
                            item.setCheckState(QtCore.Qt.Unchecked)
                    idx += 1

            if self._all_option and not self._multi:
                self.addItem("<all>")

            # Make sure names are visible even with drop down arrow
            width = self.minimumSizeHint().width()
            self.setMinimumWidth(width+50)
        finally:
            self.blockSignals(False)
        
        if self._multi:
            self.setCurrentIndex(0)
        elif isinstance(current, six.string_types):
            idx = self.findText(current)
            if idx >= 0:
                self.setCurrentIndex(idx)
            elif not self._explicit:
                # Default to first item, but do a quick switch to 
                # make sure signal is sent when first data arrives
                self.setCurrentIndex(-1)
                self.setCurrentIndex(0)
            else:
                self.setCurrentIndex(-1)
        elif self._explicit:
            self.setCurrentIndex(-1)
        else:
            # Must be <all> option
            self.setCurrentIndex(0)
        self._update_highlight()

    def setEnabled(self, enable):
        """
        Overridden from QtGui.QWidget

        Only highlight selector in red when widget is enabled
        """
        QtGui.QWidget.setEnabled(self, enable)
        self._update_highlight()

    def _update_highlight(self):
        if self._explicit and self.isEnabled() and self.currentIndex() == -1:
            self.setStyleSheet("QComboBox {background-color: #d05050}")
        else:
            self.setStyleSheet("")

class ChoiceOption(Option, QtGui.QComboBox):
    """ 
    Option which is chosen from a list of possible strings 
    """
    sig_changed = QtCore.Signal()

    def __init__(self, choices=(), return_values=None, default=None):
        QtGui.QComboBox.__init__(self)
        self.setChoices(choices, return_values)
        # Bizarre hack to make the dropdown height adjust to the items added
        self.setView(QtGui.QListView())
        if default:
            self.value = default
        self.currentIndexChanged.connect(self._changed)

    def setChoices(self, choices, return_values=None):
        """
        Set the list of options to be chosen from

        :param choices: Sequence of strings
        :param return_values: Optional matching sequence of strings to be returned 
                              as the ``value`` for each choice
        """
        if return_values is None:
            return_values = list(choices)

        if len(return_values) != len(choices):
            raise QpException("Number of return values must match number of choices")
        self.return_values = return_values
        self.choice_map = dict(zip([str(choice) for choice in choices], return_values))
        
        try:
            self.blockSignals(True)
            self.clear()
            for choice in choices:
                self.addItem(str(choice))
            self.setCurrentIndex(0)
        finally:
            self.blockSignals(False)
        self._changed()

    @property
    def value(self):
        """ 
        Value of currently selected option. 
        
        This is either the selected text, or the corresponding return value
        if these were supplied when the object was created
        """
        return self.choice_map.get(self.currentText(), None)
    
    @value.setter
    def value(self, choice):
        idx = self.findText(str(choice))
        if idx < 0:
            for ret_idx, ret in enumerate(self.return_values):
                if ret == choice:
                    idx = ret_idx
                    break
            
        if idx >= 0:
            self.setCurrentIndex(idx)
        else:
            raise ValueError("Value %s is not a valid choice for this option" % choice)

    def _changed(self):
        self.sig_changed.emit()

class TextOption(Option, QtGui.QLineEdit):
    """ 
    Option which contains arbitrary text
    """
    sig_changed = QtCore.Signal()

    def __init__(self, initial=""):
        QtGui.QLineEdit.__init__(self, initial)
        self.editingFinished.connect(self._changed)

    @property
    def value(self):
        """ Current text """
        return self.text()
    
    @value.setter
    def value(self, name):
        self.setText(name)

    def _changed(self):
        self.sig_changed.emit()
     
class OutputNameOption(TextOption):
    """ 
    Option used to specify the output data name for a process
    """
    
    def __init__(self, src_data=None, suffix="_out", initial="output"):
        TextOption.__init__(self)
        self.src_data = src_data
        self.initial = initial
        self.suffix = suffix
        self._reset()
        if src_data is not None:
            src_data.sig_changed.connect(self._reset)

    def _reset(self):
        if self.src_data is not None:
            self.setText(self.src_data.value + self.suffix)
        else:
            self.setText(self.initial)
     
class NumericOption(Option, QtGui.QWidget):
    """
    Numeric option chooser which uses a slider and two spin boxes
    """
    sig_changed = QtCore.Signal()

    def __init__(self, minval=0, maxval=100, default=0, intonly=False, decimals=2, **kwargs):
        QtGui.QWidget.__init__(self)
        self.minval = minval
        self.maxval = maxval
        self.hardmin = kwargs.get("hardmin", False)
        self.hardmax = kwargs.get("hardmax", False)
        self.valid = True

        if intonly:
            self.rtype = int
            self.decimals = 0
        else:
            self.rtype = float
            self.decimals = decimals
            
        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        #self.min_edit = QtGui.QLineEdit(str(minval))
        #self.min_edit.editingFinished.connect(self_min_changed)
        #hbox.addWidget(self.min_edit)

        self.slider = QtGui.QSlider(QtCore.Qt.Horizontal)
        self.slider.setMaximum(100)
        self.slider.setMinimum(0)
        self.slider.setSliderPosition(int(100 * (default - minval) / (maxval - minval)))
        self.slider.valueChanged.connect(self._slider_changed)
        if kwargs.get("slider", True):
            hbox.addWidget(self.slider)

        self.val_edit = QtGui.QLineEdit(str(default))
        self.val_edit.editingFinished.connect(self._edit_changed)
        hbox.addWidget(self.val_edit)

    def _changed(self):
        self.sig_changed.emit()

    def _edit_changed(self):
        try:
            val = self.rtype(self.val_edit.text())
            self.valid = True
            self.val_edit.setStyleSheet("")

            if val > self.maxval and not self.hardmax:
                self.maxval = val
            if val < self.minval and not self.hardmin:
                self.minval = val

            val = self.value
            pos = 100 * (val - self.minval) / (self.maxval - self.minval)
            try:
                self.slider.blockSignals(True)
                self.slider.setSliderPosition(int(pos))
            finally:
                self.slider.blockSignals(False)

        except ValueError:
            self.val_edit.setStyleSheet("QLineEdit {background-color: #d05050}")
            self.valid = False

        self._changed()

    def _slider_changed(self, value):
        val = self.rtype(self.minval + (self.maxval - self.minval) * float(value) / 100)
        try:
            self.val_edit.blockSignals(True)
            self._update_edit(val)
        finally:
            self.val_edit.blockSignals(False)
        self._changed()

    @property
    def value(self):
        """ The numeric value selected """
        if self.valid:
            return self.rtype(self.val_edit.text())
        else:
            raise QpException("'%s' is not a valid number")

    @value.setter
    def value(self, value):
        self._update_edit(value)
        self._edit_changed()
    
    def _update_edit(self, value):
        if self.rtype == int:
            self.val_edit.setText(str(value))
        else:
            self.val_edit.setText(sf(value, sig_fig=self.decimals))

    def setLimits(self, minval=None, maxval=None):
        """
        Set the minimum and maximum slider values
        """
        if minval:
            self.minval = minval
        if maxval:
            self.maxval = maxval

        try:
            self.blockSignals(True)
            if self.valid:
                self._edit_changed()
        finally:
            self.blockSignals(False)

class BoolOption(Option, QtGui.QCheckBox):
    """ 
    Option used to specify a true or false value
    """
    sig_changed = QtCore.Signal()

    def __init__(self, default=False, invert=False):
        """
        :param default: Initial value of ``value`` property
        :param invert: If True, ``value`` property is the opposite of the check state
        """ 
        QtGui.QCheckBox.__init__(self)
        if invert: default = not default
        self.setChecked(default)
        self.invert = invert
        self.stateChanged.connect(self._changed)

    @property
    def value(self):
        """ True or False according to whether the option is selected """
        if self.invert:
            return not self.isChecked()
        else:
            return self.isChecked()
    
    @value.setter
    def value(self, checked):
        if self.invert:
            checked = not checked
        self.setChecked(checked)

    def _changed(self):
        self.sig_changed.emit()
      
class MatrixOption(Option, QtGui.QTableView):
    """
    Option which returns a 2D matrix of numbers
    """

    sig_changed = QtCore.Signal()

    def __init__(self, initial, col_headers=None, row_headers=None, expandable=(True, True),
                 fix_height=False, fix_width=False, readonly=False):
        QtGui.QTableView.__init__(self)
        
        self._model = QtGui.QStandardItemModel()
        self.setModel(self._model)
        self.updating = False
        self.expandable = expandable
        self.row_headers = row_headers
        self.col_headers = col_headers
        self.fix_height = fix_height
        self.fix_width = fix_width
        self.default_bg = None
        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        self.setMatrix(initial, False, col_headers=col_headers, row_headers=row_headers)
        
        if readonly:
            self.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        else:
            self.model().itemChanged.connect(self._item_changed)
        
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        #self.horizontalHeader().setResizeMode(QtGui.QHeaderView.ResizeToContents)
        self.itemDelegate().closeEditor.connect(self._expand_if_required, QtCore.Qt.QueuedConnection)
        self.setAcceptDrops(True)
        
    def valid(self):
        """
        :return: True if matrix contains valid numeric values
        """
        try:
            self.values()
            return True
        except ValueError:
            return False

    @property
    def value(self):
        """ Matrix of numbers as a sequence of sequences """
        rows = []
        try:
            for r in range(self._model.rowCount()-int(self.expandable[1])):
                row = [float(self._model.item(r, c).text()) for c in range(self._model.columnCount()-int(self.expandable[0]))]
                rows.append(row)
        except:
            raise ValueError("Non-numeric data in list")
        return np.array(rows)

    @value.setter
    def value(self, mat):
        self.setMatrix(mat)

    def _set_size(self):
        # QTableWidget is completely incapable of choosing a sensible size. We do our best
        # here but have to allow a bit of random 'padding' in case scrollbars are necessary
        tx, ty = self.horizontalHeader().length()+15, self.verticalHeader().length()+15
        if self.row_headers is not None: tx += self.verticalHeader().width()
        if self.col_headers is not None: ty += self.horizontalHeader().height()
        #sh = self.sizeHint()
        #tx = min(tx, sh.width())
        #ty = min(ty, sh.height())
        if self.fix_height: self.setFixedHeight(ty)
        if self.fix_width: self.setFixedWidth(tx)

    def setMatrix(self, matrix, validate=True, col_headers=None, row_headers=None):
        """
        Set the values in the matrix
        """
        if validate:
            if not matrix:
                raise ValueError("No values provided")
            elif len(matrix[0]) != self._model.columnCount() and not self.expandable[0]:
                raise ValueError("Incorrect number of columns - expected %i" % self._model.columnCount())
            elif len(matrix) != self._model.rowCount() and not self.expandable[1]:
                raise ValueError("Incorrect number of rows - expected %i" % self._model.rowCount())

        if (col_headers and self.expandable[0]) or (row_headers and self.expandable[1]):
            raise RuntimeError("Can't specify headers for auto-expandable dimensions")

        self._model.setRowCount(len(matrix)+int(self.expandable[1]))
        self._model.setColumnCount(len(matrix[0])+int(self.expandable[0]))

        if col_headers:
            self._model.setHorizontalHeaderLabels(col_headers)
            self.horizontalHeader().show()

        if row_headers:
            self._model.setVerticalHeaderLabels(row_headers)
            self.verticalHeader().show()
        
        self._model.blockSignals(True)
        try:
            for r, rvals in enumerate(matrix):
                for c, v in enumerate(rvals):
                    item = QtGui.QStandardItem("%g" % v)
                    self._model.setItem(r, c, item)
            
            if self.expandable[0]:
                for r in range(len(matrix)):
                    item = QtGui.QStandardItem("")
                    self._model.setItem(r, self._model.columnCount()-1, item)
        
            if self.expandable[1]:
                for c in range(len(matrix[0])):
                    item = QtGui.QStandardItem("")
                    self._model.setItem(self._model.rowCount()-1, c, item)
        
            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        finally:
            self._model.blockSignals(False)
            
        self._set_size()
        self.sig_changed.emit()

    def _range_empty(self, cs, rs):
        empty = True
        for r in rs:
            for c in cs:
                # Check if r or c is negative and do not consider empty
                # so last row/column will not be deleted
                if r < 0 or c < 0 or (self._model.item(r, c) is not None and self._model.item(r, c).text() != ""):
                    empty = False
                    break
        return empty

    def _item_changed(self, item):
        if self.updating: return
        self.updating = True
        try:
            if self.default_bg is None:
                self.default_bg = item.background()

            try:
                float(item.text())
                item.setBackground(self.default_bg)
            except ValueError:
                if item.text() != "": 
                    item.setBackground(QtGui.QColor('#d05050'))

            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        finally:
            self.updating = False

    def _expand_if_required(self):
        last_col = self._model.columnCount()-1
        last_row = self._model.rowCount()-1
        if self.expandable[0]:
            if not self._range_empty([last_col], range(self._model.rowCount())):
                # Last column is not empty - add a new empty column
                self._model.setColumnCount(self._model.columnCount()+1)
                # Set rest of new column to prevent it from immediately being invalid
                for row in range(self._model.rowCount()-int(self.expandable[1])):
                    item = self._model.item(row, last_col)
                    if item is None or item.text() == "":
                        item = QtGui.QStandardItem(self._model.item(row, last_col-1).text())
                        self._model.setItem(row, last_col, item)
                    
            elif self._range_empty([self._model.columnCount()-2], range(self._model.rowCount())):
                # Last two columns are empty, so remove last column
                self._model.setColumnCount(self._model.columnCount()-1)

        if self.expandable[1]:
            if not self._range_empty(range(self._model.columnCount()), [last_row]):
               # Last row is not empty - add a new empty row
                self._model.setRowCount(self._model.rowCount()+1)
                # Set rest of new row to prevent it from immediately being invalid
                for col in range(self._model.columnCount()-int(self.expandable[0])):
                    item = self._model.item(last_row, col)
                    if item is None or item.text() == "":
                        self._model.setItem(last_row, col, QtGui.QStandardItem(self._model.item(last_row-1, col).text()))
            
            elif self._range_empty(range(self._model.columnCount()), [self._model.rowCount()-2]):
                # Last-but-one row is empty, so remove last row (which is always empty)
                self._model.setColumnCount(self._model.columnCount()-1)
        self.sig_changed.emit()

    def dragEnterEvent(self, event):
        """ Called when item is dragged into the matrix grid"""
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """ Called when item is dragged over the matrix grid"""
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """ 
        Called when item is dropped onto the matrix grid
        
        If it's a filename-like item, try to load a matrix
        from the file. Otherwise ignore it
        """
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                self._load_file(local_file_from_drop_url(url))
        else:
            event.ignore()
            
    def _load_file(self, filename):
        fvals, _, ncols = load_matrix(filename)
        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        else:
            self.setMatrix(fvals)

class VectorOption(MatrixOption):
    """
    Option which returns a list of numbers
    """
    def __init__(self, initial, row=True, headers=None, expandable=True, fix_size=True):
        self._row = row
        initial = self._to_matrix(initial)
        if row:
            row_headers, col_headers = None, headers
            expandable = (expandable, False)
            fix_height, fix_width = fix_size, False
        else:
            row_headers, col_headers = headers, None
            expandable = (False, expandable)
            fix_height, fix_width = False, fix_size

        MatrixOption.__init__(self, initial, 
                              row_headers=row_headers, col_headers=col_headers, 
                              expandable=expandable, 
                              fix_height=fix_height, fix_width=fix_width)

    def _to_matrix(self, vals):
        if self._row:
            return [vals,]
        else:
            return [v[0] for v in vals]

    def _from_matrix(self, matrix):
        if self._row:
            return matrix[0]
        else:
            return np.array([row[0] for row in matrix])

    @property
    def value(self):
        """ Sequence of numbers """
        return self._from_matrix(MatrixOption.value.fget(self))

    def setList(self, values, **kwargs):
        """
        Set the matrix values as a 1D list
        """
        MatrixOption.setMatrix(self, self._to_matrix(values), **kwargs)

    def _load_file(self, filename):
        fvals, nrows, ncols = load_matrix(filename)

        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        elif ncols == 1:
            self.setList([row[0] for row in fvals])
        elif nrows == 1:
            self.setList(fvals[0])
        else:
            # Choose row or column you want
            row, col = self._choose_row_col(fvals)
            if row is not None:
                self.setList(fvals[row])
            elif col is not None:
                vals = [v[col] for v in fvals]
                self.setList(vals)

    def _choose_row_col(self, vals):
        d = MatrixViewerDialog(self, vals, title="Choose a row or column", text="Select a row or column containing the data you want")
        if d.exec_():
            ranges = d.table.selectedRanges()
            if ranges:
                r = ranges[0]
                if r.topRow() == r.bottomRow():
                    # Row select
                    return r.topRow(), None
                elif r.leftColumn() == r.rightColumn():
                    # Column select
                    return None, r.leftColumn()
        return None, None    
  
class NumberListOption(Option, QtGui.QWidget):
    """ 
    A list of numbers which may be entered space or comma separated
    """
    sig_changed = QtCore.Signal()

    def __init__(self, initial=(), intonly=False, load_btn=True):
        QtGui.QWidget.__init__(self)

        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        self._edit = QtGui.QLineEdit()
        self._edit.dropEvent = self.dropEvent
        hbox.addWidget(self._edit)
        self._btn = QtGui.QPushButton("Load")
        self._btn.clicked.connect(self._load_clicked)
        if load_btn:
            hbox.addWidget(self._btn)

        if intonly:
            self._type = int
        else:
            self._type = float
        self.value = initial
        self._edit.editingFinished.connect(self._edit_changed)

    @property
    def value(self):
        """ List of numbers or empty list if invalid data entered """
        try:
            text = self._edit.text().replace(",", " ")
            return [self._type(v) for v in text.split()] 
        except ValueError:
            return []

    @value.setter
    def value(self, vals):
        self._edit.setText(" ".join([str(v) for v in vals]))
        self._edit_changed()

    def _edit_changed(self):
        try:
            text = self._edit.text().replace(",", " ")
            numbers = [self._type(v) for v in text.split()]
            self._edit.setText(" ".join([str(v) for v in numbers]))
            self._edit.setStyleSheet("")
        except ValueError:
            # Colour edit red but don't change anything
            self._edit.setStyleSheet("QLineEdit {background-color: #d05050}")
        self.sig_changed.emit()

    def _load_clicked(self):
        filename, _ = QtGui.QFileDialog.getOpenFileName(dir=default_save_dir())
        if filename:
            set_default_save_dir(os.path.dirname(filename))
            self._load_file(filename)

    def dragEnterEvent(self, event):
        """ Called when item is dragged into the matrix grid"""
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """ Called when item is dragged over the matrix grid"""
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """ 
        Called when item is dropped onto the matrix grid
        
        If it's a filename-like item, try to load a matrix
        from the file. Otherwise ignore it
        """
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                self._load_file(local_file_from_drop_url(url))
        else:
            event.ignore()

    def _load_file(self, filename):
        fvals, nrows, ncols = load_matrix(filename)

        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        elif ncols == 1:
            self.value = [row[0] for row in fvals]
        elif nrows == 1:
            self.value = fvals[0]
        else:
            # Choose row or column you want
            row, col = self._choose_row_col(fvals)
            if row is not None:
                self.value = fvals[row]
            elif col is not None:
                self.value = [v[col] for v in fvals]

    def _choose_row_col(self, vals):
        d = MatrixViewerDialog(self, vals, title="Choose a row or column", text="Select a row or column containing the data you want")
        if d.exec_():
            ranges = d.table.selectedRanges()
            if ranges:
                r = ranges[0]
                if r.topRow() == r.bottomRow():
                    # Row select
                    return r.topRow(), None
                elif r.leftColumn() == r.rightColumn():
                    # Column select
                    return None, r.leftColumn()
        return None, None

class PickPointOption(Option, QtGui.QWidget):
    """ 
    Option used to specify a single point in a data set
    """
    sig_changed = QtCore.Signal()

    def __init__(self, ivl, grid=None, intonly=True):
        """
        :param grid: DataGrid instance - output position will be reported relative to this grid
        :param intonly: If True, positions will be rounded to nearest integer
        """
        QtGui.QWidget.__init__(self)
        self._ivl = ivl
        self._grid = grid
        if intonly:
            self._rtype = int
            self._offset = 0.5
        else:
            self._rtype = float
            self._offset = 0

        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        self._edit = QtGui.QLineEdit()
        self._edit.editingFinished.connect(self._edit_changed)
        hbox.addWidget(self._edit)
        self._btn = QtGui.QPushButton("Pick point")
        self._btn.clicked.connect(self._pick_point)
        hbox.addWidget(self._btn)

        self._point = None
        
    @property
    def value(self):
        """ 3D position as float sequence relative to the grid specified in ``setGrid`` """
        return self._point
    
    def setGrid(self, grid):
        """
        Set the grid to be used when reporting the output point
        
        :param grid: DataGrid instance
        """
        if self._point:
            self._set_point(grid.grid_to_grid(self._point, from_grid=self._grid))
        self._grid = grid

    def _edit_changed(self):
        self._point = None
        try:
            point = [self._rtype(float(v)) for v in self._edit.text().split()]
            if len(point) == 3:
                self._point = point
        except ValueError:
            # Leave point as None if text is not valid
            pass

        self.sig_changed.emit()
      
    def _pick_point(self):
        self._ivl.set_picker(PickMode.SINGLE)
        self._ivl.sig_selection_changed.connect(self._point_picked)
        self._btn.setEnabled(False)

    def _point_picked(self, picker):
        self._set_point(picker.selection(self._grid))
        self._btn.setEnabled(True)
        self._ivl.sig_selection_changed.disconnect(self._point_picked)

    def _set_point(self, point):
        self._edit.setText(" ".join([str(self._rtype(v+self._offset)) for v in point[:3]]))
        self._edit_changed()

class FileOption(Option, QtGui.QWidget):
    """ 
    Option used to specify a file or directory
    """
    sig_changed = QtCore.Signal()

    def __init__(self, dirs=False, initial=""):
        """
        :param dirs: If True, allow only directories to be selected
        """
        QtGui.QWidget.__init__(self)
        self._dirs = dirs
        
        hbox = QtGui.QHBoxLayout()
        self.setLayout(hbox)
        self._edit = QtGui.QLineEdit(initial)
        hbox.addWidget(self._edit)
        self._btn = QtGui.QPushButton("Choose")
        self._btn.clicked.connect(self._clicked)
        hbox.addWidget(self._btn)
        
    @property
    def value(self):
        """ Return path of selected file """
        return self._edit.text()

    def _clicked(self):
        if self._dirs:
            path = QtGui.QFileDialog.getExistingDirectory(dir=self.value)
        else:
            path, _ = QtGui.QFileDialog.getOpenFileName(dir=os.path.dirname(self.value))
        self._edit.setText(path)
        self.sig_changed.emit()

class RunButton(QtGui.QWidget):
    """
    A button which, when clicked, runs an analysis process
    """

    def __init__(self, label="Run", callback=None):
        QtGui.QWidget.__init__(self)
        hbox = QtGui.QHBoxLayout()
        self.setLayout(hbox)
        self.btn = QtGui.QPushButton(label)
        self.btn.clicked.connect(callback)
        hbox.addWidget(self.btn)
        hbox.addStretch(1)



"""
Quantiphyse - Experimental widgets (not currently in GUI)

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import print_function, division

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, HelpButton

class ImageExportWidget(QpWidget):

    def __init__(self, **kwargs):
        super(ImageExportWidget, self).__init__(name="Image Export", desc="Export images and animations", icon="image_export", **kwargs)

    def init_ui(self):
        main_vbox = QtGui.QVBoxLayout()
        
        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(QtGui.QLabel('<font size="5">Export animation</font>'))
        hbox.addStretch(1)
        hbox.addWidget(HelpButton(self))
        main_vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        b1 = QtGui.QPushButton('Generate', self)
        b1.clicked.connect(self.run_time_window_capture)
        hbox.addWidget(b1)
        hbox.addStretch(1)
        main_vbox.addLayout(hbox)

        main_vbox.addStretch(1)
        self.setLayout(main_vbox)

    @QtCore.Slot()
    def run_time_window_capture(self):
        """
        Capture 4D changes over time
        """
        if self.ivm.vol is None:
            error_dialog("No data loaded")
            return

        shape = self.ivm.vol.shape

        # Choose a folder to save images
        fname = QtGui.QFileDialog.getExistingDirectory(self, 'Choose folder to save images')
        if fname == '':
            return

        for ii in range(shape[-1]):
            self.debug("Frame number:", ii)
            self.ivl.set_time_pos(ii)
            output_name = fname + '/' + str(ii).zfill(3) + '.png'
            self.ivl.capture_view_as_image(1, output_name)

QP_WIDGETS = []








"""
Quantiphyse - Subclass of QpData for handling Nifti data

Copyright (c) 2013-2018 University of Oxford
"""
from __future__ import division, print_function

import os
import logging
import traceback

import nibabel as nib
import numpy as np

from .qpdata import DataGrid, QpData, NumpyData

LOG = logging.getLogger(__name__)

QP_NIFTI_EXTENSION_CODE = 42

class NiftiData(QpData):
    """
    QpData from a Nifti file
    """
    def __init__(self, fname):
        nii = nib.load(fname)
        shape = list(nii.shape)
        while len(shape) < 3:
            shape.append(1)

        if len(shape) > 3:
            nvols = shape[3]
        else:
            nvols = 1

        self.rawdata = None
        self.voldata = None
        self.nifti_header = nii.header
        metadata = None
        for ext in self.nifti_header.extensions:
            if ext.get_code() == QP_NIFTI_EXTENSION_CODE:
                import yaml
                LOG.debug("Found QP metadata: %s", ext.get_content())
                try:
                    metadata = yaml.load(ext.get_content())[0]["QpMetadata"]
                    LOG.debug(metadata)
                except (KeyError, yaml.YAMLError):
                    LOG.warn("Failed to read Quantiphyse metadata")
                    LOG.warn(ext.get_content())
                    traceback.print_exc()

        xyz_units, vol_units = "mm", None
        units = nii.header.get_xyzt_units()
        if units:
            xyz_units = units[0]
            if len(units) > 1:
                vol_units = units[1]

        vol_scale = 1.0
        zooms = nii.header.get_zooms()
        if zooms and len(zooms) > 3:
            vol_scale = zooms[3]

        grid = DataGrid(shape[:3], nii.header.get_best_affine(), units=xyz_units)
        QpData.__init__(self, fname, grid, nvols, vol_unit=vol_units, vol_scale=vol_scale, fname=fname, metadata=metadata)

    def raw(self):
        # NB: copy() converts data to an in-memory array instead of a numpy file memmap.
        # Appears to improve speed drastically as well as stop a bug with accessing the subset of the array
        # memmap has been designed to save space on ram by keeping the array on the disk but does
        # horrible things with performance, and analysis especially when the data is on the network.
        if self.rawdata is None:
            nii = nib.load(self.fname)
            #self.rawdata = nii.get_data().copy()
            self.rawdata = nii.get_data()
            self.rawdata = self._correct_dims(self.rawdata)

        self.voldata = None
        return self.rawdata
        
    def volume(self, vol, qpdata=False):
        vol = min(vol, self.nvols-1)
        if self.nvols == 1:
            ret = self.raw()
        elif self.rawdata is not None:
            ret = self.rawdata[:, :, :, vol]
        else:
            if self.voldata is None:
                self.voldata = [None,] * self.nvols
            if self.voldata[vol] is None:
                nii = nib.load(self.fname)
                self.voldata[vol] = self._correct_dims(nii.dataobj[..., vol])
            ret = self.voldata[vol]

        if qpdata:
            return NumpyData(ret, grid=self.grid, name="%s_vol_%i" % (self.name, vol))
        else:
            return ret

    def _correct_dims(self, arr):
        while arr.ndim < 3:
            arr = np.expand_dims(arr, -1)

        if self.metadata.get("raw_2dt", False) and arr.ndim == 3:
            # Single-slice, interpret 3rd dimension as time
            arr = np.expand_dims(arr, 2)

        if arr.ndim == 4 and arr.shape[3] == 1:
            arr = np.squeeze(arr, axis=-1)
        return arr

def save(data, fname, grid=None, outdir=""):
    """
    Save data to a file
    
    :param data: QpData instance
    :param fname: File name
    :param grid: If specified, grid to save the data on
    :param outdir: Optional output directory if fname is not absolute
    """
    if grid is None:
        grid = data.grid
        arr = data.raw().copy()
    else:
        arr = data.resample(grid).raw().copy()
        
    if hasattr(data, "nifti_header"):
        header = data.nifti_header.copy()
    else:
        header = None

    img = nib.Nifti1Image(arr, grid.affine, header=header)
    img.update_header()
    if data.metadata:
        from quantiphyse.utils.batch import to_yaml
        yaml_metadata = to_yaml({"QpMetadata" : data.metadata})
        LOG.debug("Writing metadata: %s", yaml_metadata)
        extensions = nib.nifti1.Nifti1Extensions([ext for ext in img.header.extensions if ext.get_code() != QP_NIFTI_EXTENSION_CODE])
        extensions.append(nib.nifti1.Nifti1Extension(QP_NIFTI_EXTENSION_CODE, yaml_metadata.encode('utf-8')))
        img.header.extensions = extensions

    if not fname:
        fname = data.name
        
    _, extension = os.path.splitext(fname)
    if extension == "":
        fname += ".nii"
        
    if not os.path.isabs(fname):
        fname = os.path.join(outdir, fname)

    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    LOG.debug("Saving %s as %s", data.name, fname)
    img.to_filename(fname)
    data.fname = fname



"""
Quantiphyse - tests for ImageVolumeManagement

Copyright (c) 2013-2018 University of Oxford
"""

import unittest
import numpy as np

from quantiphyse.data import ImageVolumeManagement, NumpyData, DataGrid

GRIDSIZE = 5

class IVMTest(unittest.TestCase):

    def setUp(self):
        self.ivm = ImageVolumeManagement()

    def testEmptyCreate(self):
        self.assertEqual(len(self.ivm.data), 0)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertTrue(self.ivm.main is None)
        
    def testAdd(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)

    def testAddRoi(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.randint(0, 10, size=shape), name="test", grid=grid, roi=True)
        self.assertTrue(qpd.roi)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)

    def testAddTwo(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.rand(*shape), name="test2", grid=grid)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_data, qpd2)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)

    def testAddTwoRois(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.randint(0, 10, size=shape), name="test", grid=grid, roi=True)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertTrue(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 2)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_roi, qpd2)
        self.assertTrue(self.ivm.current_data is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test"], qpd)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddTwoMixed1(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_roi, qpd2)
        self.assertTrue(self.ivm.current_data is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddTwoMixed2(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd2)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd2)
        self.assertEqual(self.ivm.current_data, qpd)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddThreeMixed(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.rand(*shape), name="test2", grid=grid)
        qpd3 = NumpyData(np.random.randint(0, 10, size=shape), name="test3", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertFalse(qpd2.roi)
        self.assertTrue(qpd3.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.ivm.add(qpd3)
        self.assertEqual(len(self.ivm.data), 3)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_data, qpd2)
        self.assertEqual(self.ivm.current_roi, qpd3)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.data["test3"], qpd3)
        self.assertEqual(self.ivm.rois["test3"], qpd3)

    def testDelete(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.ivm.delete(qpd.name)
        self.assertEqual(len(self.ivm.data), 0)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertTrue(self.ivm.main is None)
      
    def testRename(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.ivm.rename("test", "test2")
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, self.ivm.data["test2"])
        self.assertTrue(np.all(self.ivm.data["test2"].raw() == qpd.raw()))

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Rag-bag of utility functions - should be moved into specific modules

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, print_function

import sys
import tempfile
import logging

from matplotlib import cm
import numpy as np
import pandas as pd

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data.extras import DataFrameExtra
from .exceptions import QpException
from .logger import LogSource
from .plugins import get_plugins
from .local import set_default_save_dir, default_save_dir, get_local_file, get_local_shlib, get_icon, set_local_file_path, local_file_from_drop_url

__all__ = ["QpException", "LogSource", "get_plugins", "get_local_file", "get_local_shlib", "get_icon", "set_local_file_path", "local_file_from_drop_url",]

DEFAULT_SIG_FIG = 4
LOG = logging.getLogger(__name__)

def norecurse(fn):
    """
    Decorator which prevents recursive calling of an object method
    """
    def _wrapper(*args):
        self = args[0]
        if not self.updating:
            try:
                self.updating = True
                return fn(*args)
            finally:
                self.updating = False
    return _wrapper

def ifnone(obj, alt):
    """
    Convenience function to return an alternative if an object is None

    Why isn't this in the standard library!

    :param obj: Object
    :param alt: Alternative
    :return: obj if not None, otherwise alt
    """
    if obj is None:
        return alt
    else:
        return obj

def show_help(section="", base='http://quantiphyse.readthedocs.io/en/latest/'):
    """
    Open the help page at a specified page

    :param section: Section ID, if not specified open main index page
    :param base: Base URL for documentation, may be version dependent
    """ 
    if section != "" and not section.endswith(".html") and not section.endswith("/"): 
        section += ".html"
    link = base + section
    QtGui.QDesktopServices.openUrl(QtCore.QUrl(link, QtCore.QUrl.TolerantMode))

def load_matrix(filename):
    """
    Load a file containing a matrix of numbers in ASCII text

    :return: Matrix of numbers (list of list, not Numpy array), number of rows, number of columns
    """
    with open(filename, "r") as matrix_file:
        return text_to_matrix(matrix_file.read())

def text_to_matrix(text):
    """
    Convert ASCII text into a matrix of numbers
    
    :return: Matrix of numbers (list of list, not Numpy array), number of rows, number of columns
    """
    fvals = []
    ncols = -1
    lines = text.splitlines()
    for line in lines:
        # Discard comments
        line = line.split("#", 1)[0].strip()
        # Split by commas or spaces
        vals = line.replace(",", " ").split()
        # Ignore empty lines
        if not vals: continue
        # Check correct number of columns
        if ncols < 0: ncols = len(vals)
        elif len(vals) != ncols:
            raise QpException("File must contain a matrix of numbers with fixed size (rows/columns)")
        # Check all data is numeric
        for val in vals:
            try:
                float(val)
            except:
                raise QpException("Non-numeric value '%s' found in matrix" % val)
        fvals.append([float(v) for v in vals])     
    
    return fvals, len(fvals), ncols

def write_temp_matrix(prefix, matrix):
    """
    Write a matrix to a temporary file, returning the filename
    """
    temp_file = tempfile.NamedTemporaryFile(prefix=prefix, delete=False)
    LOG.debug("matrix: %s", prefix)
    for row in matrix:
        line = " ".join([str(v) for v in row]) + "\n"
        LOG.debug(line)
        temp_file.write(line)
    temp_file.close()
    return temp_file.name

def get_version():
    """
    :return: Quantiphyse version number
    """
    try:
        from .._version import __version__
        return __version__
    except ImportError:
        LOG.warn(sys.exc_info()[1])
        return "<version not found>"

def table_to_extra(tabmod, name):
    """ 
    Turn a QT table model into a DataFrameExtra
    """
    cols = range(tabmod.columnCount())
    rows = range(tabmod.rowCount())
    columns = [tabmod.horizontalHeaderItem(col).text().replace("\n", " ") for col in cols]
    index = [tabmod.verticalHeaderItem(row).text() for row in rows] 

    rowdata = []
    for row in rows:
        rowdata.append([tabmod.item(row, col).text() for col in cols])

    df = pd.DataFrame(rowdata, index=index, columns=columns)
    return DataFrameExtra(name, df)

def copy_table(tabmod):
    """ Copy a QT table model to the clipboard in a form suitable for paste into Excel etc """
    clipboard = QtGui.QApplication.clipboard()
    tsv = str(table_to_extra(tabmod, ""))
    clipboard.setText(tsv)

def sf(num, sig_fig=None):
    """ Format a number as a string to a given number of sig figs """
    if sig_fig is None: 
        global DEFAULT_SIG_FIG
        sig_fig = DEFAULT_SIG_FIG
    fmt = "%%.%ig" % sig_fig
    return fmt % float(num)

def remove_nans(data, fillvalue=0):
    """
    Check for and remove nans from data arrays
    """
    notnans = np.isfinite(data)
    if not np.all(notnans):
        data[np.logical_not(notnans)] = fillvalue

def get_col(cmap, idx, out_of):
    """ Get RGB color for an index within a range, using a Matplotlib colour map """
    if out_of == 0: 
        return [255, 0, 0]
    else:
        return [int(255 * rgbf) for rgbf in cmap(float(idx)/out_of)[:3]]

def get_pencol(roi, region):
    """
    Get an RGB pen colour for a given ROI region
    """
    if roi is None:
        return (255, 0, 0)
    else:
        return get_lut(roi)[region]

def get_lut(roi, alpha=None):
    """
    Get the colour look up table for the ROI.
    """
    cmap = getattr(cm, 'jet')
    try:
        max_region = max(roi.regions.keys())
    except:
        # No nonzero regions!
        max_region = 1
    if max_region < 3: max_region = 3
    lut = [[0, 0, 0],] + [[int(255 * rgb1) for rgb1 in cmap(float(v+1)/max_region)[:3]] for v in range(max_region-1, -1, -1)]
    lut = np.array(lut, dtype=np.ubyte)

    if alpha is not None:
        # add transparency
        alpha1 = np.ones((lut.shape[0], 1))
        alpha1 *= alpha
        alpha1[0] = 0
        lut = np.hstack((lut, alpha1))

    return lut

# Kelly (1965) - set of 20 contrasting colours
# We alter the order a bit to prioritize those that give good contrast to our dark background
# plus we add an 'off white' at the start
KELLY_COLORS = [
    ("off_white", (230, 230, 230)),
    ("vivid_yellow", (255, 179, 0)),
    ("vivid_orange", (255, 104, 0)),
    ("very_light_blue", (166, 189, 215)),
    ("vivid_red", (193, 0, 32)),
    ("grayish_yellow", (206, 162, 98)),
    ("medium_gray", (129, 112, 102)),
    ("strong_purple", (128, 62, 117)),

    # these aren't good for people with defective color vision:
    ("vivid_green", (0, 125, 52)),
    ("strong_purplish_pink", (246, 118, 142)),
    ("strong_blue", (0, 83, 138)),
    ("strong_yellowish_pink", (255, 122, 92)),
    ("strong_violet", (83, 55, 122)),
    ("vivid_orange_yellow", (255, 142, 0)),
    ("strong_purplish_red", (179, 40, 81)),
    ("vivid_greenish_yellow", (244, 200, 0)),
    ("strong_reddish_brown", (127, 24, 13)),
    ("vivid_yellowish_green", (147, 170, 0)),
    ("deep_yellowish_brown", (89, 51, 21)),
    ("vivid_reddish_orange", (241, 58, 19)),
    ("dark_olive_green", (35, 44, 22)),
]

def get_kelly_col(idx, wrap=True):
    """
    Get the Kelly colour for a given index

    :param idx: Index
    :param wrap: If True, wrap index to number of Kelly colours
    :return: (RGB) tuple in range 0-255
    """
    baseidx = idx % len(KELLY_COLORS)
    basecol = KELLY_COLORS[baseidx][1]
    if not wrap:
        raise RuntimeError("Not implemented yet")
    return basecol



"""
Main program entry point for Quantiphyse

Copyright (c) 2013-2018 University of Oxford
"""
import sys
import warnings
if "--debug" not in sys.argv:
    # Avoid ugly warnings from some third party packages unless we are debugging
    warnings.simplefilter("ignore")

import argparse
import multiprocessing
import signal
import traceback
import logging

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
    PYSIDE1 = True
except ImportError:
    # Note that pyqtgraph actually writes all the contents of QtWidgets into
    # QtGui on import! This is sort-of nice because we don't need to switch
    # existing PySide code that uses, e.g. QtGui.QMainWindow, but it's a bit
    # invasive compared with the 'nicer' option or importing PySide.QtGui
    # as QtWidgets. We will go with the pyqtgraph method for now but might
    # need to make changes if this causes problems later.
    from PySide2 import QtGui, QtCore, QtWidgets
    PYSIDE1 = False

from quantiphyse.test import run_tests
from quantiphyse.utils import QpException, set_local_file_path
from quantiphyse.utils.batch import BatchScript
from quantiphyse.utils.logger import set_base_log_level
from quantiphyse.utils.local import get_icon
from quantiphyse.gui import MainWindow, Register
from quantiphyse.gui.dialogs import error_dialog, set_main_window

# Required to use resources in theme. Check if 2 or 3.
if sys.version_info[0] > 2:
    from .resources import resource_py3
else:
    from .resources import resource_py2

def my_catch_exceptions(exc_type, exc, tb):
    """
    Catch exceptions and format appropriately

    QpException can occur due to bad user input so scary tracebacks are not included.
    Other exception types are bugs so give full traceback
    """
    if issubclass(exc_type, QpException):
        detail = exc.detail
    else:
        detail = traceback.format_exception(exc_type, exc, tb)
    error_dialog(str(exc), title="Error", detail=detail)

def main():
    """
    Parse any input arguments and run the application
    """

    # Enable multiprocessing on windows frozen binaries. Does nothing on other systems
    multiprocessing.freeze_support()

    # Parse input arguments to pass info to GUI
    parser = argparse.ArgumentParser()
    parser.add_argument('data', help='Load data files', nargs="*", type=str)
    parser.add_argument('--batch', help='Run batch file', default=None, type=str)
    parser.add_argument('--debug', help='Activate debug mode', action="store_true")
    parser.add_argument('--test-all', help='Run all tests', action="store_true")
    parser.add_argument('--test', help='Specify test suite to be run (default=run all)', default=None)
    parser.add_argument('--test-fast', help='Run only fast tests', action="store_true")
    parser.add_argument('--qv', help='Activate quick-view mode', action="store_true")
    parser.add_argument('--register', help='Force display of registration dialog', action="store_true")
    args = parser.parse_args()

    # Apply global options
    if args.debug:
        set_base_log_level(logging.DEBUG)
    else:
        set_base_log_level(logging.WARN)

    if args.register:
        QtCore.QSettings().setValue("license_accepted", 0)

    # Set the local file path, used for finding icons, plugins, etc
    set_local_file_path()

    # Handle CTRL-C correctly
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Check whether any batch processing arguments have been called
    if args.test_all or args.test:
        app = QtCore.QCoreApplication(sys.argv)
        run_tests(args.test)
        sys.exit(0)
    elif args.batch is not None:
        app = QtCore.QCoreApplication(sys.argv)
        runner = BatchScript()
        # Add delay to make sure script is run after the main loop starts, in case
        # batch script is completely synchronous
        QtCore.QTimer.singleShot(200, lambda: runner.execute({"yaml-file" : args.batch}))
    else:
        # OS specific changes
        if sys.platform.startswith("darwin") and PYSIDE1:
            QtGui.QApplication.setGraphicsSystem('native')

        app = QtGui.QApplication(sys.argv)
        app.setStyle('plastique')
        QtCore.QCoreApplication.setOrganizationName("ibme-qubic")
        QtCore.QCoreApplication.setOrganizationDomain("eng.ox.ac.uk")
        QtCore.QCoreApplication.setApplicationName("Quantiphyse")
        QtGui.QApplication.setWindowIcon(QtGui.QIcon(get_icon("main_icon.png")))

        if args.debug:
            import pyqtgraph as pg
            pg.systemInfo()

        # Create window and start main loop
        pixmap = QtGui.QPixmap(get_icon("quantiphyse_splash.png"))
        splash = QtGui.QSplashScreen(pixmap)
        splash.show()
        app.processEvents()

        win = MainWindow(load_data=args.data, widgets=not args.qv)
        splash.finish(win)
        sys.excepthook = my_catch_exceptions
        set_main_window(win)
        Register.check_register()
    sys.exit(app.exec_())



"""
Quantiphyse - PCA reduction process

Copyright (c) 2013-2018 University of Oxford
"""
import six

from quantiphyse.data.extras import MatrixExtra
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

import numpy as np

class HistogramProcess(Process):
    """
    Calculate histogram for a data set 
    """
    
    PROCESS_NAME = "Histogram"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data_items = options.pop('data', None)
        if data_items is None:
            data_items = self.ivm.data.keys()
        elif isinstance(data_items, six.string_types):
            data_items = [data_items,]

        if not data_items:
            raise QpException("No data to calculate histogram")
        
        data_items = [self.ivm.data[name] for name in data_items]
        roi = self.get_roi(options, use_current=False)
        
        sel_region = options.pop('region', None)
        dmin = options.pop('min', None)
        dmax = options.pop('max', None)
        bins = options.pop('bins', 100)
        vol = options.pop('vol', None)
        prob = options.pop('yscale', 'count').lower().startswith("prob")
        output_name = options.pop('output-name', "histogram")

        yvals, col_headers = {}, ["left", "right", "centre",]
        for data in data_items:
            if vol is not None:
                rawdata = data.volume(vol)
            else:
                rawdata = data.raw()
            hrange = [dmin, dmax]
            if dmin is None: hrange[0] = np.min(rawdata)
            if dmax is None: hrange[1] = np.max(rawdata)

            if roi is None:
                roi_fordata = np.ones(data.grid.shape)
                regions = {1 : ""}
            else:
                roi_fordata = roi.resample(data.grid).raw()
                regions = roi.regions

            for region, region_name in regions.items():
                if sel_region is not None and region != sel_region:
                    self.debug("Ignoring region %i", region)
                    continue

                region_data = rawdata[roi_fordata == region]
                data_vals, edges = np.histogram(region_data, bins=bins, range=hrange, density=prob)
                    
                if region_name:
                    name = "%s\n%s" % (data.name, region_name)
                else:
                    name = data.name

                col_headers.append(name)
                yvals[name] = data_vals
            
        xvals = [[edges[idx], edges[idx+1], (edges[idx]+edges[idx+1])/2] for idx in range(len(edges)-1)]
        rows = []
        for idx, row in enumerate(xvals):
            for name in col_headers[3:]:
                row += [yvals[name][idx],]
            rows.append(row)

        if not options.pop('no-extras', False):
            self.debug("Adding %s" % output_name)
            extra = MatrixExtra(output_name, rows, col_headers=col_headers)
            self.debug(str(extra))
            self.ivm.add_extra(output_name, extra)



"""
Quantiphyse - Widgets for comparing data sets

Copyright (c) 2013-2018 University of Oxford
"""

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
import pyqtgraph as pg
import numpy as np

from quantiphyse.gui.widgets import QpWidget, TitleWidget, OverlayCombo, NumericOption, ChoiceOption
from quantiphyse.utils import QpException

class CompareDataWidget(QpWidget):
    """
    Compare two data sets
    """
    def __init__(self, **kwargs):
        QpWidget.__init__(self, name="Compare Data", icon="compare.png", 
                          desc="Compare two data sets", group="Visualisation", **kwargs)
        
    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, batch_btn=False, help="compare")
        vbox.addWidget(title)

        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(QtGui.QLabel("Compare "))
        self.d1_combo = OverlayCombo(self.ivm)
        self.d1_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.d1_combo)
        hbox.addWidget(QtGui.QLabel(" with "))
        self.d2_combo = OverlayCombo(self.ivm)
        self.d2_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.d2_combo)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(QtGui.QLabel("Within ROI "))
        self.roi_combo = OverlayCombo(self.ivm, rois=True, data=False, none_option=True)
        self.roi_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.roi_combo)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        gbox = QtGui.QGroupBox("Options")
        grid = QtGui.QGridLayout()
        gbox.setLayout(grid)

        self.id_cb = QtGui.QCheckBox("Include identity line")
        grid.addWidget(self.id_cb, 0, 0)
        self.sample_cb = QtGui.QCheckBox("Sample values")
        self.sample_cb.setChecked(True)
        self.sample_cb.stateChanged.connect(self._update_data)
        self.sample_cb.stateChanged.connect(self._update_gui)
        grid.addWidget(self.sample_cb, 1, 0)
        self.sample_spin = QtGui.QSpinBox()
        self.sample_spin.setMinimum(10)
        self.sample_spin.setMaximum(10000000)
        self.sample_spin.setSingleStep(100)
        self.sample_spin.setValue(1000)
        self.sample_spin.valueChanged.connect(self._update_data)
        grid.addWidget(self.sample_spin, 1, 1)
        self.warn_label = QtGui.QLabel("WARNING: plotting all values may take a long time")
        self.warn_label.setStyleSheet("QLabel { color : red; }")
        self.warn_label.setVisible(False)
        grid.addWidget(self.warn_label, 2, 0)
        self.plot_mode = ChoiceOption("Plot mode", grid, 3, choices=["Scatter", "Heat map"])
        self.plot_mode.combo.currentIndexChanged.connect(self._plot_mode_changed)
        self.bins = NumericOption("Bins", grid, 3, xpos=2, minval=20, default=50, intonly=True)
        self.bins.label.setVisible(False)
        self.bins.spin.setVisible(False)
        self.run_btn = QtGui.QPushButton("Update")
        self.run_btn.clicked.connect(self._run)
        grid.addWidget(self.run_btn, 4, 0)
        hbox.addWidget(gbox)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        win = pg.GraphicsLayoutWidget()
        win.setBackground(background=None)
        self.plot = win.addPlot(enableAutoRange=True)
        self.plot.clear() # Unsure why this is necessary
        hbox.addWidget(win)
        
        self.hist = pg.HistogramLUTWidget()
        self.hist.setVisible(False)
        self.hist.setBackground(None)
        self.hist.gradient.loadPreset("thermal")
        hbox.addWidget(self.hist) 
        vbox.addLayout(hbox)

        self._update_gui()
        self._update_data()
    
    def _plot_mode_changed(self, idx):
        self.bins.label.setVisible(idx == 1)
        self.bins.spin.setVisible(idx == 1)
        self.sample_cb.setChecked(idx == 0)
        self._update_gui()
        self._update_data()

    def _update_gui(self):
        self.sample_spin.setEnabled(self.sample_cb.isChecked())

    def _update_data(self):
        name1 = self.d1_combo.currentText()
        name2 = self.d2_combo.currentText()
        roi_name = self.roi_combo.currentText()
        qpd1 = self.ivm.data.get(name1, None)
        qpd2 = self.ivm.data.get(name2, None)
        roi = self.ivm.rois.get(roi_name, None)

        if qpd1 is not None and qpd2 is not None:
            current_vol = self.ivl.focus()[3]
            d1 = qpd1.volume(current_vol)
            d2 = qpd2.resample(qpd1.grid).volume(current_vol)
            
            if roi is not None:
                roi_data = roi.resample(qpd1.grid).raw()
                d1 = d1[roi_data > 0]
                d2 = d2[roi_data > 0]
            else:
                d1 = d1.reshape(-1)
                d2 = d2.reshape(-1)

            if self.sample_cb.isChecked():
                n_samples = self.sample_spin.value()
                idx = np.random.choice(np.arange(len(d1)), n_samples)
                d1 = np.take(d1, idx)
                d2 = np.take(d2, idx)
                self.warn_label.setVisible(False)
            else:
                # Warn about plotting all data in scatter mode
                self.warn_label.setVisible(self.plot_mode.combo.currentIndex() == 0)
                        
            self.d1 = d1
            self.d2 = d2
            self.run_btn.setEnabled(True)
        else:
            self.run_btn.setEnabled(False)

    def _run(self):
        self.plot.clear() 
        self.plot.setLabel('bottom', self.d1_combo.currentText())
        self.plot.setLabel('left', self.d2_combo.currentText())
        if self.plot_mode.combo.currentIndex() == 0:
            self.plot.setAspectLocked(False)
            self.hist.setVisible(False)
            self.plot.plot(self.d1, self.d2, pen=None, symbolBrush=(200, 200, 200), symbolPen='k', symbolSize=5.0)
        else:
            heatmap, xedges, yedges = np.histogram2d(self.d1, self.d2, bins=self.bins.value())
            rect = QtCore.QRectF(xedges[0], yedges[0], xedges[-1]-xedges[0], yedges[-1]-yedges[0])
            img = pg.ImageItem(heatmap)
            img.setRect(rect)
            #self.plot.setAspectLocked(True)
            self.hist.setVisible(True)
            self.hist.setImageItem(img)
            self.plot.addItem(img)

        if self.id_cb.isChecked():
            real_min = max(self.d1.min(), self.d2.min())
            real_max = min(self.d1.max(), self.d2.max())
            pen = pg.mkPen((255, 255, 255), style=QtCore.Qt.DashLine)
            self.plot.plot([real_min, real_max], [real_min, real_max], pen=pen, width=2.0)



"""
Tests for Radial profile widget - empty at present

Copyright (c) 2013-2018 University of Oxford
"""
import unittest

from quantiphyse.test.widget_test import WidgetTest

from .widget import RadialProfileWidget

class RadialProfileWidgetTest(WidgetTest):

    def widget_class(self):
        return RadialProfileWidget

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - tests for QpData classes

Copyright (c) 2013-2018 University of Oxford
"""

import os
import unittest
import tempfile

import numpy as np

from quantiphyse.data import NumpyData, DataGrid
import quantiphyse.data.nifti as nifti

GRIDSIZE = 5
NVOLS = 4

class NumpyDataTest(unittest.TestCase):
    """ Tests for the NumpyData subclass of QpData """

    def setUp(self):
        self.shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        self.grid = DataGrid(self.shape, np.identity(4))
        self.floats = np.random.rand(*self.shape)
        self.ints = np.random.randint(0, 10, self.shape)
        self.floats4d = np.random.rand(*(self.shape + [NVOLS,]))

    def test3d(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, 1)
        self.assertEqual(qpd.ndim, 3)
        self.assertFalse(qpd.roi)
        
    def test3dints(self):
        qpd = NumpyData(self.ints, grid=self.grid, name="test", roi=True)
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, 1)
        self.assertEqual(qpd.ndim, 3)
        self.assertTrue(qpd.roi)
        
    def test4d(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, NVOLS)
        self.assertEqual(qpd.ndim, 4)
        self.assertFalse(qpd.roi)
        
    def testRandomName(self):
        name = "sdkfjhsdfi4"
        qpd = NumpyData(self.floats4d, grid=self.grid, name=name)
        self.assertEqual(qpd.name, name)
        
    def testRoiFloats(self):
        """ Check that ROIs can contain float data so long as the numbers are really integers """
        qpd = NumpyData(self.ints.astype(np.float), grid=self.grid, name="test", roi=True)
        self.assertTrue(issubclass(qpd.raw().dtype.type, np.floating))
        self.assertTrue(qpd.roi)
        
    def testRoiRegions(self):
        qpd = NumpyData(self.ints, grid=self.grid, name="test", roi=True)
        self.assertTrue(qpd.roi)
        regions = [v for v in np.unique(self.ints) if v != 0]
        self.assertEqual(regions, list(qpd.regions.keys()))

    def testRange(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        mx, mn = np.max(self.floats), np.min(self.floats)
        self.assertAlmostEqual(qpd.range()[0], mn)
        self.assertAlmostEqual(qpd.range()[1], mx)
        
    def testSet2dt(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        qpd.set_2dt()
        self.assertEqual(qpd.nvols, GRIDSIZE)
        self.assertEqual(qpd.ndim, 4)
        d = qpd.raw()
        for idx in range(GRIDSIZE):
            vol = d[..., idx]
            self.assertEqual(vol.shape[0], GRIDSIZE)
            self.assertEqual(vol.shape[1],  GRIDSIZE)
            self.assertEqual(vol.shape[2], 1)
            self.assertTrue(np.allclose(np.squeeze(vol), self.floats[:, :, idx]))

    def testVolume(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        for idx in range(NVOLS):
            self.assertTrue(np.allclose(qpd.volume(idx), self.floats4d[..., idx]))
        
    def testValue3d(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        POS = [2, 3, 4]
        self.assertAlmostEqual(qpd.value(POS), self.floats[POS[0], POS[1], POS[2]])
           
    def testValue3d4dPos(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        POS = [2, 3, 4, 1]
        self.assertAlmostEqual(qpd.value(POS), self.floats[POS[0], POS[1], POS[2]])
     
    def testValue4d(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        POS = [2, 3, 4, 1]
        self.assertAlmostEqual(qpd.value(POS), self.floats4d[POS[0], POS[1], POS[2], POS[3]])
        
    def testValue4d3dPos(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        POS = [2, 3, 4]
        self.assertAlmostEqual(qpd.value(POS), self.floats4d[POS[0], POS[1], POS[2], 0])
        
class NiftiDataTest(unittest.TestCase):
    """ Tests for the NiftiData subclass of QpData """

    def setUp(self):
        self.shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        self.grid = DataGrid(self.shape, np.identity(4))
        self.floats = np.random.rand(*self.shape)
        self.ints = np.random.randint(0, 10, self.shape)
        self.floats4d = np.random.rand(*(self.shape + [NVOLS,]))

    def testSaveLoadMetadata(self):
        tempdir = tempfile.mkdtemp(prefix="qp")
        fname = os.path.join(tempdir, "test.nii.gz")
        md = {
            "flibble" : 2.3,
            "flobble" : 4,
            "fishcakes" : ["a", "b", "c"],
            "fluntycaps" : {
                "gfruffle" : "def",
                "gfable" : 8,
            }
        }
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test", metadata=md)
        nifti.save(qpd, fname)

        nifti_data = nifti.NiftiData(fname)
        for key, value in md.items():
            self.assertTrue(key in nifti_data.metadata)
            self.assertEqual(nifti_data.metadata[key], value)

    def testLoadSaveSameName(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")

        tempdir = tempfile.mkdtemp(prefix="qp")
        fname = os.path.join(tempdir, "test.nii")
        nifti.save(qpd, fname)

        nifti_data = nifti.NiftiData(fname)
        nifti.save(nifti_data, fname)

if __name__ == '__main__':
    unittest.main()



import unittest

from quantiphyse.utils import get_plugins

from .ivm_test import IVMTest
from .qpd_test import NumpyDataTest, NiftiDataTest
from .slice_plane_test import OrthoSliceTest
from .io_test import IoProcessTest

class_tests = [IVMTest, NumpyDataTest, NiftiDataTest, OrthoSliceTest, IoProcessTest,]

def run_tests(test_filter=None):
    """
    Run all unit tests defined by packages and plugins

    :param test_filter: Specifies name of test set to be run, None=run all
    """
    suite = unittest.TestSuite()

    for test in class_tests:
        if test_filter is None or test.__name__.lower().startswith(test_filter.lower()):
            suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(test))

    tests = get_plugins("widget-tests")
    for test in tests:
        if test_filter is None or test.__name__.lower().startswith(test_filter.lower()):
            suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(test))

    tests = get_plugins("process-tests")
    for test in tests:
        if test_filter is None or test.__name__.lower().startswith(test_filter.lower()):
            suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(test))
   
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Quantiphyse - Widgets for data registration and motion correction

Copyright (c) 2013-2018 University of Oxford
"""
import traceback

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import QpData
from quantiphyse.gui.widgets import QpWidget, RunWidget, TitleWidget, RunButton
from quantiphyse.gui.options import OptionBox, Option, DataOption, ChoiceOption, NumericOption, TextOption, OutputNameOption
from quantiphyse.utils import get_plugins

class RegWidget(QpWidget):
    """
    Generic registration / motion correction widget 
    """
    def __init__(self, **kwargs):
        super(RegWidget, self).__init__(name="Registration", icon="reg", 
                                        desc="Registration and Motion Correction", 
                                        group="Registration", **kwargs)
        self.reg_methods = []
        for method in get_plugins("reg-methods"):
            try:
                self.reg_methods.append(method(self.ivm))
            except:
                traceback.print_exc()
                self.warn("Failed to create registration method: %s", method)
                
    def init_ui(self):
        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        title = TitleWidget(self, title="Registration and Motion Correction", help="reg")
        layout.addWidget(title)

        if not self.reg_methods:
            layout.addWidget(QtGui.QLabel("No registration methods found"))
            layout.addStretch(1)
            return

        self.options = OptionBox("General Options")
        self.options.add("Mode", ChoiceOption(["Registration", "Motion Correction"], ["reg", "moco"]), key="mode")
        self.options.add("Method", ChoiceOption([method.display_name for method in self.reg_methods], self.reg_methods), key="method")
        self.options.add("Registration data", DataOption(self.ivm), key="reg")
        self.options.add("Reference data", DataOption(self.ivm), key="ref")
        self.options.add("Reference volume", ChoiceOption(["Middle volume", "Mean volume", "Specified volume"], ["median", "mean", "idx"]), key="ref-vol")
        self.options.add("Reference volume index", NumericOption(intonly=True), key="ref-idx")
        self.options.add("Output space", ChoiceOption(["Reference", "Registration", "Transformed"], ["ref", "reg", "trans"]), key="output-space")
        self.options.add("Output name", OutputNameOption(src_data=self.options.option("reg"), suffix="_reg"), key="output-name", checked=True)
        self.options.add("Also apply transform to", DataOption(self.ivm, multi=True), key="add-reg")
        self.options.add("Save transformation", TextOption(), key="save-transform", checked=True, default=False)
        
        self.options.option("mode").sig_changed.connect(self._update_option_visibility)
        self.options.option("method").sig_changed.connect(self._method_changed)
        self.options.option("ref").sig_changed.connect(self._update_option_visibility)
        self.options.option("ref-vol").sig_changed.connect(self._update_option_visibility)
        self.options.option("reg").sig_changed.connect(self._update_option_visibility)
        layout.addWidget(self.options)

        # Create the options boxes for reg methods - only one visible at a time!
        self.opt_boxes = {}
        for method in self.reg_methods:
            hbox = QtGui.QHBoxLayout()
            opt_box = QtGui.QGroupBox()
            opt_box.setTitle(method.display_name)
            vbox = QtGui.QVBoxLayout()
            opt_box.setLayout(vbox)
            vbox.addWidget(method.interface())
            hbox.addWidget(opt_box)
            opt_box.setVisible(False)
            layout.addLayout(hbox)
            self.opt_boxes[method.name] = opt_box

        layout.addWidget(RunWidget(self))
        layout.addStretch(1)
        self._method_changed()

    def _method_changed(self):
        method = self.options.option("method").value
        for name, box in self.opt_boxes.items():
            box.setVisible(name == method.name)
        self.options.option("save-transform").value = "%s_trans" % method.name
        self._update_option_visibility()

    def _update_option_visibility(self):
        mode = self.options.option("mode").value
        regdata = self.ivm.data.get(self.options.option("reg").value, None)
        refdata = self.ivm.data.get(self.options.option("ref").value, None)
        refvol = self.options.option("ref-vol").value

        nvols_reg, nvols_ref = 1, 1
        if regdata is not None:
            nvols_reg = regdata.nvols
            
        if mode == "moco" and regdata is not None:
            nvols_ref = regdata.nvols
        elif mode == "reg" and refdata is not None:
            nvols_ref = refdata.nvols

        self.options.set_visible("ref", mode == "reg")
        self.options.set_visible("ref-vol", nvols_ref > 1)
        self.options.set_visible("ref-idx", nvols_ref > 1 and refvol == "idx")
        self.options.set_visible("add-reg", nvols_reg == 1 and mode == "reg")
        self.options.set_visible("output-space", mode == "reg")
        
        if nvols_ref > 1:
            self.options.option("ref-idx").setLimits(0, nvols_ref-1)
            self.options.option("ref-idx").value = int(nvols_ref/2)

    def processes(self):
        options = self.options.values()
        if options.get("ref-vol", None) == "idx":
            options["ref-vol"] = options.pop("ref-idx")
        
        method = options.pop("method")
        options["method"] = method.name
        options.update(method.options())

        return {
            "Reg" : options,
        }

class TransformOption(Option, QtGui.QComboBox):
    """ 
    Option for choosing previously calculated registration transforms. 
    These may be stored as data sets or Extras
    """
    sig_changed = QtCore.Signal()

    def __init__(self, ivm):
        QtGui.QComboBox.__init__(self)
        self.ivm = ivm
        self._data_changed()
        self.ivm.sig_all_data.connect(self._data_changed)
        self.ivm.sig_extras.connect(self._data_changed)
        self.currentIndexChanged.connect(self._changed)

    @property
    def value(self):
        """ 
        Name of currently selected transform
        """
        return self.currentText()

    def _data_changed(self):
        self.blockSignals(True)
        try:
            current = self.value
            self.clear()

            idx, current_idx = 0, 0
            items = list(self.ivm.data.values()) + list(self.ivm.extras.values())
            for item in items:
                if "QpReg" in item.metadata:
                    self.addItem(item.name)
                    if item.name == current:
                        current_idx = idx
                    idx += 1

            # Make sure names are visible even with drop down arrow
            width = self.minimumSizeHint().width()
            self.setMinimumWidth(width+50)
        finally:
            self.blockSignals(False)
        
        # Set to same item selected beforehand and make sure
        # sig_changed is always emitted
        self.setCurrentIndex(current_idx)
        self._changed()
       
    def _changed(self):
        self.sig_changed.emit()

class TransformDetails(QtGui.QGroupBox):
    """
    Widget displaying information about a selected transformation
    """

    def __init__(self):
        QtGui.QGroupBox.__init__(self, "Transform details")
        grid = QtGui.QGridLayout()
        self.setLayout(grid)
        
        self._name = QtGui.QLineEdit()
        grid.addWidget(QtGui.QLabel("Transform name"), 0, 0)
        grid.addWidget(self._name, 0, 1)

        self._method = QtGui.QLineEdit()
        grid.addWidget(QtGui.QLabel("Registration method"), 1, 0)
        grid.addWidget(self._method, 1, 1)

        self._type = QtGui.QLineEdit()
        grid.addWidget(QtGui.QLabel("Transform type"), 2, 0)
        grid.addWidget(self._type, 2, 1)

        self._edit = QtGui.QPlainTextEdit()
        self._edit.setReadOnly(True)
        grid.addWidget(self._edit, 3, 0, 1, 2)

    @property
    def transform(self):
        """ The transform object """
        return self._transform

    @transform.setter
    def transform(self, transform):
        self._transform = transform
        self._name.setText(self._transform.name)
        self._method.setText(self._transform.metadata.get("QpReg", "<unknown>"))
        if isinstance(transform, QpData):
            self._type.setText("Image (e.g. warp field)")
            self._edit.setVisible(False)
        else:
            self._type.setText("Non-image (e.g. matrix)")
            self._edit.setVisible(True)
            doc = self._edit.document()
            font = doc.defaultFont()
            font.setFamily("Courier New")
            font.setPointSize(8)
            doc.setDefaultFont(font)
            self._edit.setPlainText(str(transform))

class ApplyTransform(QpWidget):
    """
    Widget for applying previously calculated transformations
    """
    def __init__(self, **kwargs):
        super(ApplyTransform, self).__init__(name="Apply Transform", icon="reg", 
                                             desc="Apply previously calculated transformations", 
                                             group="Registration", **kwargs)
        self.reg_methods = []
        for method in get_plugins("reg-methods"):
            try:
                self.reg_methods.append(method(self.ivm))
            except:
                traceback.print_exc()
                self.warn("Failed to create registration method: %s", method)

    def init_ui(self):
        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        title = TitleWidget(self, help="reg")
        layout.addWidget(title)

        if not self.reg_methods:
            layout.addWidget(QtGui.QLabel("No registration methods found"))
            layout.addStretch(1)
            return

        self.options = OptionBox("General Options")
        self.options.add("Transform", TransformOption(self.ivm), key="transform")
        self.options.add("Apply to data", DataOption(self.ivm), key="data")
        self.options.add("Interpolation", ChoiceOption(["Nearest neighbour", "Linear", "Spline"], [0, 1, 3], default=1), key="interp-order")
        self.options.add("Output name", OutputNameOption(src_data=self.options.option("data"), suffix="_reg"), key="output-name")
        self.options.option("transform").sig_changed.connect(self._transform_changed)
        layout.addWidget(self.options)

        self.details = TransformDetails()
        layout.addWidget(self.details)

        layout.addWidget(RunButton(self))
        layout.addStretch(1)
        self._transform_changed()

    def processes(self):
        return {
            "ApplyTransform" : self.options.values(),
        }

    def activate(self):
        self._transform_changed()

    def _transform_changed(self):
        trans_name = self.options.option("transform").value
        transform = self.ivm.data.get(trans_name, None)
        if transform is None or "QpReg" not in transform.metadata:
            transform = self.ivm.extras.get(trans_name, None)

        if transform is not None and "QpReg" in transform.metadata:
            self.details.transform = transform



"""
Quantiphyse - Widgets which perform data clustering

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import numpy as np
from sklearn.metrics import pairwise

import pyqtgraph as pg
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData
from quantiphyse.gui.widgets import QpWidget, OverlayCombo, RoiCombo, NumericOption, TitleWidget
from quantiphyse.utils import get_pencol

from .kmeans import KMeansProcess, MeanValuesProcess

class ClusteringWidget(QpWidget):
    """
    Widget for doing K-means clustering on 3D or 4D data
    """

    def __init__(self, **kwargs):
        super(ClusteringWidget, self).__init__(name="KMeans Clustering", icon="clustering", desc="Generate clusters from 3D or 4D data", group="Clustering", position=2, **kwargs)
        self.curves = {}

    def init_ui(self):
        self.process = KMeansProcess(self.ivm)
        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        title = TitleWidget(self, help="cluster")
        layout.addWidget(title)
        
        DESC = """
<i>Performs clustering of 3D or 4D data using the K-means algorithm.
PCA reduction is used on 4D data to extract representative curves
"""
        desc = QtGui.QLabel(DESC)
        desc.setWordWrap(True)
        layout.addWidget(desc)

        gbox = QtGui.QGroupBox()
        gbox.setTitle('Clustering options')

        grid = QtGui.QGridLayout()
        gbox.setLayout(grid)

        # Data to cluster
        grid.addWidget(QtGui.QLabel("Data"), 0, 0)
        self.data_combo = OverlayCombo(self.ivm)
        self.data_combo.currentIndexChanged.connect(self._data_changed)
        grid.addWidget(self.data_combo, 0, 1)

        grid.addWidget(QtGui.QLabel("ROI"), 1, 0)
        self.roi_combo = RoiCombo(self.ivm, none_option=True)
        grid.addWidget(self.roi_combo, 1, 1)

        # Number of clusters inside the ROI
        self.n_clusters = NumericOption("Number of clusters", grid, xpos=2, ypos=0, minval=2, maxval=20, default=4, intonly=True)
        self.n_clusters.spin.setToolTip("")

        # Number of PCA modes
        self.n_pca = NumericOption("Number of PCA modes", grid, xpos=2, ypos=1, minval=1, maxval=10, default=3, intonly=True)
        self.n_pca.spin.setToolTip("")

        # Output ROI name
        grid.addWidget(QtGui.QLabel("Output name"), 2, 0)
        self.output_name = QtGui.QLineEdit("clusters")
        grid.addWidget(self.output_name, 2, 1)
        layout.addWidget(gbox)

        # Run clustering button
        hbox = QtGui.QHBoxLayout()
        self.run_btn = QtGui.QPushButton('Run', self)
        self.run_btn.clicked.connect(self.run_clustering)
        hbox.addWidget(self.run_btn)
        hbox.addStretch(1)
        layout.addLayout(hbox)

        # Plot window, showing representative curves for 4D data
        self.show_curves_btn = QtGui.QPushButton('Show representative curves', self)
        self.show_curves_btn.clicked.connect(self._show_curves)
        layout.addWidget(self.show_curves_btn)
        self.plotwin = pg.GraphicsLayoutWidget()
        self.plotwin.setBackground(background=None)
        self.plot = self.plotwin.addPlot(title="Cluster representative curves")
        self.plot.setLabel('left', "Signal Enhancement")
        self.plotwin.setVisible(False)        
        layout.addWidget(self.plotwin)

        # Statistics
        self.show_count_btn = QtGui.QPushButton('Show voxel counts', self)
        self.show_count_btn.clicked.connect(self._show_counts)
        layout.addWidget(self.show_count_btn)
        self.stats_gbox = QtGui.QGroupBox()
        self.stats_gbox.setTitle('Voxel count')

        self.count_table = QtGui.QStandardItemModel()
        self.count_view = QtGui.QTableView()
        self.count_view.resizeColumnsToContents()
        self.count_view.setModel(self.count_table)

        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(self.count_view)
        self.stats_gbox.setLayout(hbox)
        self.stats_gbox.setVisible(False)
        layout.addWidget(self.stats_gbox)

        # Merge regions
        self.show_merge_btn = QtGui.QPushButton('Show merge options', self)
        self.show_merge_btn.clicked.connect(self._show_merge)
        layout.addWidget(self.show_merge_btn)
        
        self.merge_gbox = QtGui.QGroupBox()
        self.merge_gbox.setTitle('Merge regions')
        vbox = QtGui.QVBoxLayout()
        self.merge_gbox.setLayout(vbox)

        hbox = QtGui.QHBoxLayout()
        self.merge_btn = QtGui.QPushButton('Merge', self)
        self.merge_btn.clicked.connect(self._run_merge)
        hbox.addWidget(self.merge_btn)
        hbox.addWidget(QtGui.QLabel('Merge region '))
        self.merge_region1 = QtGui.QLineEdit('1', self)
        hbox.addWidget(self.merge_region1)
        hbox.addWidget(QtGui.QLabel(' with '))
        self.merge_region2 = QtGui.QLineEdit('2', self)
        hbox.addWidget(self.merge_region2)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        self.auto_merge_btn = QtGui.QPushButton('AutoMerge', self)
        self.auto_merge_btn.clicked.connect(self._run_automerge)
        hbox.addWidget(self.auto_merge_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        self.merge_gbox.setVisible(False)
        layout.addWidget(self.merge_gbox)

        layout.addStretch(1)

    def activate(self):
        self.ivl.sig_focus_changed.connect(self._focus_changed)
        self.ivm.sig_current_roi.connect(self._current_roi_changed)
        self.ivm.sig_main_data.connect(self._main_data_changed)
        self._data_changed()

    def deactivate(self):
        self.ivl.sig_focus_changed.disconnect(self._focus_changed)
        self.ivm.sig_current_roi.disconnect(self._current_roi_changed)
        self.ivm.sig_main_data.disconnect(self._main_data_changed)

    def _current_roi_changed(self, roi):
        if roi is not None and roi.name != self.output_name.text():
            self.roi_combo.setCurrentIndex(self.roi_combo.findText(roi.name))
        
    def _main_data_changed(self, data):
        if data is not None:
            idx = self.data_combo.findText(data.name)
        else:
            idx = 0
        self.data_combo.setCurrentIndex(idx)

    def _data_changed(self):
        data = self.ivm.data.get(self.data_combo.currentText(), None)
        if data is not None:
            is4d = data.nvols > 1
            self.debug("Number of vols: %i, 4d=%s", data.nvols, is4d)
            self.n_pca.label.setVisible(is4d)
            self.n_pca.spin.setVisible(is4d)
            self.show_curves_btn.setVisible(is4d)
            self.plotwin.setVisible(is4d and self.plotwin.isVisible())
            self.auto_merge_btn.setEnabled(is4d)
        self.run_btn.setEnabled(data is not None)

    def _focus_changed(self):
        self._update_voxel_count()

    def _show_curves(self):
        if self.plotwin.isVisible():
            self.plotwin.setVisible(False)
            self.show_curves_btn.setText("Show representative curves")
        else:
            self.plotwin.setVisible(True)
            self.show_curves_btn.setText("Hide representative curves")
            
    def _show_counts(self):
        if self.stats_gbox.isVisible():
            self.stats_gbox.setVisible(False)
            self.show_count_btn.setText("Show voxel counts")
        else:
            self.stats_gbox.setVisible(True)
            self.show_count_btn.setText("Hide voxel counts")
        
    def _show_merge(self):
        if self.merge_gbox.isVisible():
            self.merge_gbox.setVisible(False)
            self.show_merge_btn.setText("Show merge_options")
        else:
            self.merge_gbox.setVisible(True)
            self.show_merge_btn.setText("Hide merge options")
        
    def batch_options(self):
        options = {
            "data" : self.data_combo.currentText(),
            "roi" :  self.roi_combo.currentText(),
            "n-clusters" : self.n_clusters.spin.value(),
            "output-name" : self.output_name.text(),
            "invert-roi" : False,
        }

        if options["roi"] == "<none>":
            del options["roi"]

        if self.n_pca.label.isVisible():
            # 4D PCA options
            options["n-pca"] = self.n_pca.spin.value()
            options["reduction"] = "pca"
            options["norm-data"] = True

        return "KMeans", options

    def run_clustering(self):
        """
        Run kmeans clustering using normalised PCA modes
        """
        options = self.batch_options()[1]

        self.process.run(options)
        self._update_voxel_count()
        if self.n_pca.label.isVisible():
            self.update_plot()     

    def _update_voxel_count(self):
        self.count_table.clear()
        self.count_table.setVerticalHeaderItem(0, QtGui.QStandardItem("Voxel count"))

        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None:
            col_idx = 0
            for region, name in roi.regions.items():
                self.count_table.setHorizontalHeaderItem(col_idx, QtGui.QStandardItem(name))

                # Volume count
                voxel_count = np.sum(roi.raw() == region)
                self.count_table.setItem(0, col_idx, QtGui.QStandardItem(str(np.around(voxel_count))))
                col_idx += 1

    def update_plot(self):
        """
        Plot the cluster curves
        :return:
        """
        # Clear graph
        self.plot.clear()
        self.plot.setLabel('bottom', self.opts.t_type, units=self.opts.t_unit)
        if self.plot.legend is not None:
            # Work around pyqtgraph legend bug - the legend is recreated multiple times!
            # https://stackoverflow.com/questions/42792858/pyqtgraph-delete-persisting-legend-in-pyqt4-gui
            self.plot.legend.scene().removeItem(self.plot.legend)
        data = self.ivm.data.get(self.data_combo.currentText(), None)
        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None and data is not None and data.nvols > 1:
            # Generate mean curve for each cluster
            self._generate_cluster_means(roi, data)

            self.plot.addLegend()

            # Plotting using single or multiple plots
            for region, name in roi.regions.items():
                if np.sum(self.curves[region]) == 0:
                    continue

                pencol = get_pencol(roi, region)
                curve = self.plot.plot(pen=pencol, width=8.0, name=name)
                xx = np.arange(len(self.curves[region]))
                curve.setData(xx, self.curves[region])

    def _run_merge(self):
        m1 = int(self.merge_region1.text())
        m2 = int(self.merge_region2.text())
        self._merge(m1, m2)

    def _run_automerge(self):
        # Use PCA features or true curves?
        # Mean features from each cluster
        # Distance matrix between features
        if len(self.curves) < 2: return

        curvemat = np.zeros((len(self.curves), self.ivm.main.nvols))
        row_region = {}
        idx = 0
        for region, curve in self.curves.items():
            row_region[idx] = region
            curvemat[idx, :] = curve
            idx += 1
        distmat = pairwise.euclidean_distances(curvemat)
        distmat[distmat == 0] = np.inf
        loc1 = np.where(distmat == distmat.min())[0]
        self._merge(row_region[loc1[0]], row_region[loc1[1]])

    def _generate_cluster_means(self, roi, data):
        """
        Generate the mean curves for each cluster
        Returns:
        """
        self.curves = {}
        for region in roi.regions:
            roi_data = data.mask(roi, region=region, output_flat=True)
            mean = np.median(roi_data, axis=0)
            self.curves[region] = mean

    def _merge(self, m1, m2):
        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None:
            roi_data = roi.raw()
            roi_data[roi_data == m1] = m2

            # signal the change
            self.ivm.add(NumpyData(roi_data, grid=roi.grid, name=self.output_name.text(), roi=True), make_current=True)

            # replot
            self.update_plot()
            self._update_voxel_count()

class MeanValuesWidget(QpWidget):
    """
    Convert a data + multi-level ROI into mean values data set
    """
    def __init__(self, **kwargs):
        super(MeanValuesWidget, self).__init__(name="Mean in ROI", icon="meanvals", 
                                               desc="Replace data with its mean value within each ROI region", 
                                               group="ROIs", **kwargs)

        layout = QtGui.QVBoxLayout()

        title = TitleWidget(self, "Generate Mean Values Data", help="mean_values")
        layout.addWidget(title)
        
        desc = QtGui.QLabel("This widget will convert the current data set into a "
                            "new data set in which each ROI region contains the mean "
                            "value for that region.\n\nThis is generally only useful for "
                            "multi-level ROIs such as clusters or supervoxels")
        desc.setWordWrap(True)
        layout.addWidget(desc)

        hbox = QtGui.QHBoxLayout()
        gbox = QtGui.QGroupBox()
        gbox.setTitle("Options")
        grid = QtGui.QGridLayout()
        gbox.setLayout(grid)
        
        grid.addWidget(QtGui.QLabel("Data"), 0, 0)
        self.ovl = OverlayCombo(self.ivm)
        self.ovl.currentIndexChanged.connect(self._data_changed)
        grid.addWidget(self.ovl, 0, 1)
        grid.addWidget(QtGui.QLabel("ROI regions"), 1, 0)
        self.roi = RoiCombo(self.ivm)
        grid.addWidget(self.roi, 1, 1)
        grid.addWidget(QtGui.QLabel("Output name"), 2, 0)
        self.output_name = QtGui.QLineEdit()
        grid.addWidget(self.output_name, 2, 1)

        btn = QtGui.QPushButton('Generate', self)
        btn.clicked.connect(self._generate)
        grid.addWidget(btn, 2, 0)
        hbox.addWidget(gbox)
        hbox.addStretch(1)
        layout.addLayout(hbox)
        layout.addStretch(1)
        self.setLayout(layout)

    def _data_changed(self):
        name = self.ovl.currentText()
        if name:
            self.output_name.setText(name + "_means")

    def batch_options(self):
        options = {
            "roi" : self.roi.currentText(),
            "data" : self.ovl.currentText(),
            "output-name" :  self.output_name.text()
        }
        return "MeanValues", options

    def _generate(self):
        options = self.batch_options()[1]

        if not options["data"]:
            QtGui.QMessageBox.warning(self, "No data selected", "Load data to generate mean values from", QtGui.QMessageBox.Close)
            return
        if not options["roi"]:
            QtGui.QMessageBox.warning(self, "No ROI selected", "Load an ROI for mean value regions", QtGui.QMessageBox.Close)
            return
        
        process = MeanValuesProcess(self.ivm)
        process.run(options)



from .process import PcaProcess
from .widget import PcaWidget
from .tests import PcaWidgetTest
    
QP_MANIFEST = {
    "widgets" : [PcaWidget,],
    "processes" : [PcaProcess,],
    "widget-tests" : [PcaWidgetTest,],
}



"""
Quantiphyse - Data management framework

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division

import logging
import keyword
import re

try:
    from PySide import QtCore
except ImportError:
    from PySide2 import QtCore

import numpy as np

from quantiphyse.utils import QpException

from .qpdata import QpData
from .load_save import NumpyData
from .extras import Extra

LOG = logging.getLogger(__name__)

class ImageVolumeManagement(QtCore.QObject):
    """
    Holds all image datas used in analysis

    Has to inherit from a Qt base class that supports signals

    Attributes
    ----------

      ``main`` 'main' QpData - typically viewed as a greyscale background
      ``data`` Mapping from data set name to QpData object
      ``current_data`` QpData which is the 'current' overlay data set
      ``current_roi`` QpData with ``roi=True`` used as the current ROI
      ``extras`` Mapping from name to object for miscellaneous extra data.
                 Extras must support string-conversion for writing to files.
    """
    # Signals

    # Change to main data
    sig_main_data = QtCore.Signal(object)

    # Change to current data
    sig_current_data = QtCore.Signal(object)

    # Change to set of data (e.g. new one added)
    sig_all_data = QtCore.Signal(list)

    # Change to current ROI
    sig_current_roi = QtCore.Signal(object)

    # Change to set of extras (e.g. new one added)
    sig_extras = QtCore.Signal(list)

    def __init__(self):
        super(ImageVolumeManagement, self).__init__()
        self.reset()

    def reset(self):
        """ Clear all data """
        self.main = None
        self.data = {}
        self.current_data = None
        self.current_roi = None
        self.extras = {}

        self.sig_main_data.emit(None)
        self.sig_current_data.emit(None)
        self.sig_current_roi.emit(None)
        self.sig_all_data.emit([])

    @property
    def rois(self):
        """
        :return: Dictionary of name : QpData for all data items which can be used as ROIs
        """
        return dict([(data.name, data) for data in self.data.values() if data.roi])

    def suggest_name(self, name, ensure_unique=True):
        """
        Suggest a name for new data that is suitable for use as a Python variable.
        If required, ensure name does not clash with existing names
        """
        # Remove invalid characters and make sure there's something left
        name = re.sub('[^0-9a-zA-Z_]', '', name)
        if not name: name = "data"

        # Add underscore if it's a keyword or starts with a digit
        if name[0] in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9') or keyword.iskeyword(name):
            name = "_" + name

        # Make it unique
        num = 1
        test_name = name
        while 1:
            if not ensure_unique or (test_name not in self.data):
                break
            num += 1
            test_name = "%s_%i" % (name, num)
        return test_name

    def _valid_name(self, name):
        if name is None or not re.match(r'[a-z_]\w*$', name, re.I) or keyword.iskeyword(name):
            raise QpException("'%s' is not a valid name" % name)

    def set_main_data(self, name):
        """
        Set the named data item as the main data

        :param name: Name of main data. Must be name of data item in IVM
        """
        self._data_exists(name)
        self.main = self.data[name]
        self.sig_main_data.emit(self.main)

    def add(self, data, name=None, grid=None, make_current=None, make_main=None, roi=None):
        """
        Add data item to IVM

        Data will be made the main data if make_main is True, or there is no main data
        
        Data will be made current if make_current is not specified and there is no
        current data, and make_main is not True

        :param data: Numpy array of QpData instance
        :param name: Name which must be a valid name not already used in this IVM
        :param grid: If data is a Numpy array, a DataGrid instance defining the orientation
        :param make_current: If True, make this the current data item
        :param make_main: If True, make this the main data.
        :param roi: If providing Numpy array, optionally specifies whether the data is an ROI or not
        """
        if isinstance(data, np.ndarray):
            if grid is None or name is None:
                raise RuntimeError("add: Numpy data must have a name and a grid")
            data = NumpyData(data, grid, name, roi=roi)       
        elif not isinstance(data, QpData):
            raise QpException("add: data must be Numpy array or QpData")

        if name is not None:
            data.name = name

        self._valid_name(data.name)
        
        # If replacing existing data, delete the old one first
        if data.name in self.data:
            del self.data[data.name]
            if self.current_data is not None and self.current_data.name == data.name:
                make_current = True
            if self.main is not None and self.main.name ==  data.name:
                make_main = True
            
        self.data[data.name] = data

        # Make main data if requested or if not specified and there is no current main data
        if make_main is None:
            make_main = self.main is None

        if make_main:
            self.set_main_data(data.name)

        # Emit the 'data changed' signal
        self.sig_all_data.emit(list(self.data.keys()))

        # Make current if requested, or if not specified and it is the first non-main data/ROI
        if make_current is None:
            make_current = ((data.roi and self.current_roi is None) or (not data.roi and self.current_data is None)) and not make_main
        if make_current:
            if data.roi:
                self.set_current_roi(data.name)
            else:
                self.set_current_data(data.name)

    def _data_exists(self, name):
        if name not in self.data:
            raise RuntimeError("Data '%s' does not exist" % name)

    def _roi_exists(self, name):
        if name not in self.rois:
            raise RuntimeError("ROI '%s' does not exist" % name)

    def is_main_data(self, qpd):
        """
        :param qpd: QpData instance
        :return: True if qpd is the main data
        """
        return self.main is not None and qpd is not None and self.main.name == qpd.name

    def is_current_data(self, qpd):
        """
        :param qpd: QpData instance
        :return: True if qpd is the current data
        """
        return self.current_data is not None and qpd is not None and self.current_data.name == qpd.name

    def is_current_roi(self, roi):
        """
        :param qpd: QpData instance
        :return: True if qpd is the current ROI
        """
        return self.current_roi is not None and roi is not None and self.current_roi.name == roi.name
        
    def set_current_data(self, name):
        """
        Set a named data item as the current data

        :param name: Name of data item which must exist within the IVM
        """
        if name is not None:
            self._data_exists(name)
            self.current_data = self.data[name]
        else:
            self.current_data = None
        self.sig_current_data.emit(self.current_data)

    def set_current(self, name):
        """
        Set a named data item as current (ROI or data as required)
        """
        if name is not None:
            self._data_exists(name)
            if self.data[name].roi:
                self.set_current_roi(name)
            else:
                self.set_current_data(name)
        else:
            self.set_current_data(None)
            self.set_current_roi(None)

    def rename(self, name, newname):
        """
        Rename a data item

        :param name: Name of data item which must exist within the IVM
        :param newname: New name for this data item
        """
        self._data_exists(name)
        qpd = self.data[name]
        qpd.name = newname
        del self.data[name]
        self.data[newname] = qpd
        self.sig_all_data.emit(list(self.data.keys()))

    def delete(self, name):
        """
        Delete a data item

        :param name: Name of data item which must exist within the IVM
        """
        self._data_exists(name)
        del self.data[name]
        if self.current_data is not None and self.current_data.name == name:
            self.current_data = None
            self.sig_current_data.emit(None)
        if self.main is not None and self.main.name == name:
            self.main = None
            self.sig_main_data.emit(None)
        self.sig_all_data.emit(list(self.data.keys()))

    def set_current_roi(self, name):
        """
        Set a named ROI as the current ROI

        :param name: Name of ROI which must exist within the IVM
        """
        if name is not None:
            self._roi_exists(name)
            self.current_roi = self.rois[name]
        else:
            self.current_roi = None
        self.sig_current_roi.emit(self.current_roi)

    def add_extra(self, name, obj):
        """
        Add an 'extra', which can be any result of a process which
        is not voxel data, e.g. a number, table, etc.

        Extras are required to support str() conversion so they
        can be written to a file

        :param name: Name to give the extra. If an extra already exists with this name
                     it will be overwritten
        :param obj: instance of Extra which should support str() conversion
        """
        if not isinstance(obj, Extra):
            raise ValueError("extra object must be subclasses of Extra")
        self.extras[name] = obj
        self.sig_extras.emit(self.extras.values())

    def values(self, pos, grid=None):
        """
        Get all the 3D data values at the current position

        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :return: Dictionary of data name : value
        """
        values = {}

        # loop over all loaded data and save values in a dictionary
        for name, qpd in self.data.items():
            if qpd.nvols == 1:
                values[name] = qpd.value(pos, grid)
                
        return values

    def timeseries(self, pos, grid=None):
        """
        Return time/volume series curves for all 4D data items
       
        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :return: Dictionary of data name : sequence of values
        """
        timeseries = {}
        for qpd in self.data.values():
            if qpd.nvols > 1:
                timeseries[qpd.name] = qpd.timeseries(pos, grid)
                
        return timeseries



"""
Quantiphyse - Package for data registration and motion correction

Copyright (c) 2013-2018 University of Oxford
"""
from .widget import RegWidget, ApplyTransform
from .process import RegProcess, MocoProcess, ApplyTransformProcess
from .reg_method import RegMethod

QP_MANIFEST = {
    "widgets" : [RegWidget, ApplyTransform],
    "processes" : [RegProcess, MocoProcess, ApplyTransformProcess],
    "base-classes" : [RegMethod,],
}



"""
Quantiphyse - Analysis processes for data smoothing

Copyright (c) 2013-2018 University of Oxford
"""

import numpy as np
import scipy.ndimage.filters

from quantiphyse.processes import Process
from quantiphyse.data import NumpyData

class SmoothingProcess(Process):
    """
    Simple process for Gaussian smoothing
    """
    PROCESS_NAME = "Smooth"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)

        output_name = options.pop("output-name", "%s_smoothed" % data.name)
        #kernel = options.pop("kernel", "gaussian")
        order = options.pop("order", 0)
        mode = options.pop("boundary-mode", "reflect")
        sigma = options.pop("sigma", 1.0)

        # Sigma is in mm so scale with data voxel sizes
        if isinstance(sigma, (int, float)):
            sigmas = [float(sigma) / size for size in data.grid.spacing]
        else:
            sigmas = [float(sig) / size for sig, size in zip(sigma, data.grid.spacing)]

        # Smooth multiple volumes independently
        if data.nvols > 1:
            sigmas += [0, ]

        #output = scipy.ndimage.filters.gaussian_filter(data.raw(), sigmas, order=order, mode=mode)
        output = self._norm_conv(data.raw(), sigmas, order=order, mode=mode)
        self.ivm.add(NumpyData(output, grid=data.grid, name=output_name), make_current=True)

    def _norm_conv(self, data, sigma, **kwargs):
      """
      Normalized convolution

      This is a way to compensate for data having nan/infinite values.
      Taken from stackoverflow.com/questions/18697532/gaussian-filtering-a-image-with-nan-in-python
      """
      v = data.copy()
      v[~np.isfinite(data)] = 0
      vv = scipy.ndimage.filters.gaussian_filter(v, sigma, **kwargs)

      w = 0*data.copy()+1
      w[~np.isfinite(data)] = 0
      ww = scipy.ndimage.filters.gaussian_filter(w, sigma, **kwargs)
      
      return vv/ww



"""
Quantiphyse - Histogram widget

Copyright (c) 2013-2018 University of Oxford
"""
from .widget import RadialProfileWidget
from .process import RadialProfileProcess
from .process_tests import RadialProfileProcessTest

QP_MANIFEST = {
    "widgets" : [RadialProfileWidget],
    "processes" : [RadialProfileProcess],
    "process-tests" : [RadialProfileProcessTest],
}



#from .process import SmoothingProcess
from .widget import CompareDataWidget
from .tests import CompareDataWidgetTest

QP_MANIFEST = {
    "widgets" : [CompareDataWidget,],
    "widget-tests" : [CompareDataWidgetTest],
    "processes" : []
}



"""
Quantiphyse - Generic analysis process

This module defines the ``Process`` class which is the basis of most
data processing tasks. A process takes a dictionary of options and
acts on an ``ImageVolumeManagement`` object which contains data items.

By defining a task as a Process, it becomes accessible to the batch
system. A GUI interface can then be created for the Process as a
QpWidget.

Copyright (c) 2013-2018 University of Oxford
"""

import os
import multiprocessing
import multiprocessing.pool
import threading
import traceback
import logging
import re
from six.moves import queue as singleproc_queue

import numpy as np
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData, save
from quantiphyse.utils import LogSource, QpException, get_plugins, set_local_file_path

#: Axis to split along when splitting up data sets for multiprocessing
#: Could be 0, 1 or 2, but 0 is probably optimal for Numpy arrays which are column-major by default
SPLIT_AXIS = 0

#: Whether to use multiprocessing - can be disabled for debugging
MULTIPROC = True

# Guard against processes which fail with massive logfiles
MAX_LOG_SIZE=100000

LOG = logging.getLogger(__name__)

def _worker_initialize():
    """
    Initializer function for multiprocessing workers.
    
    This makes sure plugins are loaded and paths to local files are set
    """
    set_local_file_path()
    get_plugins()

class Process(QtCore.QObject, LogSource):
    """
    A data processing task
    
    The purpose of this class and it subclasses is to expose processing tasks to the batch system, 
    and also allow them to be used from the GUI

    Processes take options in the form of a key/value dictionary derived from the YAML batch file.
    Certain option names are standardized:

      name - Optional name given to the process. Will be used to name the log file output. 
             This is set in the constructor by the batch code but is taken from the process
             options list if present. If not given as an option, the process generic name
             (e.g. MoCo) is used.
      data - The name of the main input data set
      roi - The name of the main input ROI if applicable
      output-name - The name of the output data set or ROI 

    Processes are implemented by subclassing Process and overriding ``run``. 'Fast' 
    processes such as loading/saving data and very simple image processing tasks 
    should carry out their data processing in this method. Failure should be
    signalled by raising an exception - ``QpException`` for 'expected' failures related
    to invalid user options, any other exception would imply a bug.
    
    Processes which may take time can be run as background processes. To do this, the
    ``run()`` method should end with a call to ``start_bg`` which will start the background
    workers. This method supports simple parallel execution of suitable data processing tasks.

    Background processes may override the ``finished()`` method to do any necessary post-processing
    after all workers are completed. This typically consists of getting the output back from
    the worker process(es), recombining it if required, and adding it to the IVM.

    Background process may also override the ``timeout()`` method which will be called every
    second during execution. Typically this is used to monitor the workers and emit
    ``sig_progress``.

    ``sig_finished`` is always emitted when a process completes, whether synchronously or
    asynchronously. ``sig_progress`` is always emitted with a value of 1 when a process completes 
    successfully.

    ``sig_log`` is emitted when a message is added to the log using the ``log`` method. It
    can be used to show an updating log from the process.
    
    Attributes:

      indir - Input data folder if process needs to load data
      outdir - Output data folder if process needs to save data
      status - Current process status. Subclasses should *not* set this attribute themselves.
      exception - If the process failed, this attribute contains the exception object. Subclasses
                  should *not* set this attribute themselves, they should simply raise the exception.
                  or pass it back from a worker.
    """

    #: Signal which may be emitted to track progress 
    #: Argument should be between 0 and 1 and indicate degree of completion
    sig_progress = QtCore.Signal(float)

    #: Signal which will be emitted when process finishes
    #: Arguments: (status, log, if status not SUCCEEDED or CANCELLED, exception, otherwise object())
    sig_finished = QtCore.Signal(int, str, object)

    #: Signal which may be emitted when the process starts a new step
    #: Argument is text description
    sig_step = QtCore.Signal(str)

    #: Signal which is emitted when a message is added to the log
    #: Argument is the message
    sig_log = QtCore.Signal(str)

    NOTSTARTED = 0
    RUNNING = 1
    FAILED = 2
    SUCCEEDED = 3
    CANCELLED = 4

    def __init__(self, ivm, **kwargs):
        """
        :param ivm: ImageVolumeManagement object
        
        Keyword arguments:

        :param proc_id: ID string for this process
        :param indir: Input data folder
        :param outdir: Output data folder
        :param worker_fn: For background processes a worker function
                          to call which will do the processing. This 
                          function should take parameters:
                          ``id``, ``queue``,  and a sequence of arguments.
                          It should return ``id``, True/False ``success`` and
                          an output object. If ``success=False`` the output
                          object should be an exception. Otherwise it can
                          be any pickleable object (e.g. Numpy array)
        """
        QtCore.QObject.__init__(self)
        LogSource.__init__(self)
        self.ivm = ivm
        self.proc_id = kwargs.pop("proc_id", None)
        self.indir = kwargs.pop("indir", "")
        self.outdir = kwargs.pop("outdir", "")
            
        self._log = ""
        self.status = Process.NOTSTARTED
        self._completed = False
        # We seem to get a segfault when emitting a signal with a None object
        self.exception = object()

        # Multiprocessing initialization
        self._multiproc = MULTIPROC and kwargs.get("multiproc", True)
        self._worker_fn = kwargs.get("worker_fn", None)
        self._sync = kwargs.get("sync", False)
        self._timer = None
        self._workers = []
        self._pool = None
        self._worker_output = []
        self._queue = None

    def execute(self, options):
        """
        Execute the process.

        This method should **not** be overridden. It wraps the ``run()`` method
        which should be reimplemented to do your processing. The differences
        between calling ``execute()`` and calling ``run()`` are:

         - ``execute`` will never throw an exception. Instead it will set the
           status of the process to ``FAILED`` and set the ``exception`` 
           attribute.
         - ``execute`` will ensure that the ``status`` attribute is set and
           ``sig_finished`` is called for synchronous processes, or for
           asynchronous processes which fail on startup.

        In general calling ``run()`` directly is preferred when widgets
        call their synchronous processes as it enables exceptions to be handled
        more naturally (by catching or allowing the default exception handler
        to catch them). Since asynchronous processes need to be able to 
        handle exceptions in ``sig_finished``, they may prefer to 
        call ``execute()`` instead so all the error handling can be in
        the ``sig_finished`` handler.

        :param options: Dictionary of process options
        """
        self.debug("Executing %s", self.proc_id)
        self.status = self.NOTSTARTED
        self._log = ""
        self.exception = object()
        self._completed = False
        try:
            self.run(options)
            if self.status == self.NOTSTARTED:
                self.status = self.SUCCEEDED
        except Exception as exc:
            self.status = self.FAILED
            self.exception = exc
            if self.debug_enabled():
                traceback.print_exc()

        if self.status != self.RUNNING and not self._completed:
            # Synchronous process already finished. Note that it might
            # call _complete itself in some cases (e.g. batch script)
            self.debug("Sync process done - completing")
            self._complete()
        else:
            self.debug("Async process running - will wait")

    def get_data(self, options, multi=False):
        """ 
        Standard method to get the data object the process is to operate on 
        
        If no 'data' option is specified, go with main data if it exists
        If 'multi' then allow data to be a list of items

        :param options: Dictionary of options - ``data`` will be consumed if present
        :return: QpData instance
        """
        data_name = options.pop("data", None)
        if data_name is None:
            if self.ivm.main is None:
                raise QpException("No data loaded")
            data = self.ivm.main
        elif multi and isinstance(data_name, list):
            # Allow specifying a list of data volumes which are concatenated
            if not data_name:
                raise QpException("Empty list given for data")
            for name in data_name:
                if name not in self.ivm.data:
                    raise QpException("Data not found: %s" % name)

            multi_data = [self.ivm.data[name] for name in data_name]
            nvols = sum([d.nvols for d in multi_data])
            self.debug("Multivol: nvols=%i", nvols)
            grid = None
            num_vols = 0
            for data_item in multi_data:
                if grid is None:
                    grid = data_item.grid
                    data = np.zeros(list(grid.shape) + [nvols,])
                data_item = data_item.resample(grid)
                if data_item.nvols == 1:
                    rawdata = np.expand_dims(data_item.raw(), 3)
                else:
                    rawdata = data_item.raw()
                data[..., num_vols:num_vols+data_item.nvols] = rawdata
                num_vols += data_item.nvols
            data = NumpyData(data, grid=grid, name="multi_data")
        else:
            if data_name in self.ivm.data:
                data = self.ivm.data[data_name]
            else:
                raise QpException("Data not found: %s" % data_name)
        return data

    def get_roi(self, options, grid=None, use_current=False):
        """
        Standard method to get the ROI the process is to operate on

        If no 'roi' option is specified, go with currently selected ROI, 
        if it exists.

        :param options: Dictionary of options - ``roi`` will be consumed if present
        :param grid:    If specified, return ROI on this grid
        :param use_current: If True, return current ROI if no ROI specified
        :return:        QpData instance. If no ROI can be found and ``grid`` is specified, will
                        return an ROI where all voxels are unmasked.
        """
        roi_name = options.pop("roi", None)
        if roi_name is None or roi_name.strip() == "":
            if use_current and self.ivm.current_roi is not None:
                roidata = self.ivm.current_roi
            elif grid is not None:
                roidata = NumpyData(np.ones(grid.shape[:3]), grid=grid, name="dummy_roi", roi=True)
            else:
                return None
        else:
            if roi_name in self.ivm.rois:
                roidata = self.ivm.rois[roi_name]
            else:
                raise QpException("ROI not found: %s" % roi_name)

        if grid is not None:
            roidata = roidata.resample(grid)
        return roidata

    def run(self, options):
        """ 
        Override to run the process 

        :param options: Dictionary of string : value for process options
        """
        raise NotImplementedError("Process subclasses must override `run`")

    def log(self, msg):
        """
        Add text to the log and emit sig_log
        """
        def _apply_backspaces(s):
            while True:
                # if you find a character followed by a backspace, remove both
                t = re.sub('.\b', '', s, count=1)
                if len(s) == len(t):
                    # now remove any backspaces from beginning of string
                    return re.sub('\b+', '', t)
                s = t

        self._log = _apply_backspaces(self._log + msg)
        self.sig_log.emit(msg)

    def get_log(self):
        """
        Get the process log string
        """
        return self._log

    def output_data_items(self):
        """
        Optional method allowing a Process to indicate what data items it produced after completion

        :return: a sequence of data item names that were output
        """
        return []

    def start_bg(self, args, n_workers=1):
        """
        Start a set of background workers
        
        This would normally called by ``run()`` after setting up the arguments to pass to the 
        worker run function.

        :param args: Sequence of arguments to the worker run function. All must be pickleable objects
        """
        # Only for background processes
        self._pool, self._queue = self._init_multiproc(n_workers)
        
        worker_args = self.split_args(n_workers, args)
        self._worker_output = [None, ] * n_workers
        self.status = Process.RUNNING

        if self._multiproc:
            self._workers = []
            for i in range(n_workers):
                self.debug("Starting task %i/%s...", i+1, n_workers)
                proc = self._pool.apply_async(self._worker_fn, worker_args[i], callback=self._worker_finished_cb)
                self._workers.append(proc)
            
            if self._sync:
                self.debug("Running background task synchronously")
                for i in range(n_workers):
                    self._workers[i].get()
            else:
                self._restart_timer()
        else:
            for i in range(n_workers):
                result = self._worker_fn(*worker_args[i])
                self.timeout(self._queue)
                if QtGui.qApp is not None: QtGui.qApp.processEvents()
                self._worker_finished_cb(result)
                if self.status != Process.RUNNING: 
                    break

    def _init_multiproc(self, num_tasks):
        if self._multiproc:
            LOG.debug("Initializing multiprocessing")
            queue = multiprocessing.Manager().Queue()
            pool_size = min(num_tasks, multiprocessing.cpu_count())
            pool = multiprocessing.Pool(pool_size, initializer=_worker_initialize)
        else:
            LOG.debug("Not using multiprocessing")
            queue = singleproc_queue.Queue()
            pool = None
        return pool, queue

    def cancel(self):
        """
        Cancel all workers. The status will be CANCELLED unless it is already complete
        """
        if self.status == Process.RUNNING:
            self.status = Process.CANCELLED
            if self._multiproc:
                # FIXME this does not work. Not incredibly harmful because
                # with status set to CANCELLED results are ignored anyway.
                # But workers will continue to work. Maybe look into
                # abortable_worker solution?
                pass
                #for p in self._workers:
                #    p.terminate()
                #    p.join(timeout=1.0)
            else:
                # Just setting the status is enough - no more workers
                # will be started
                pass

        self._complete()

    def timeout(self, queue):
        """
        Called every 1s while the process is running. 
        
        Override to monitor progress of job via the queue and emit 
        sig_progress / sig_step as required
        """
        pass

    def finished(self, worker_output):
        """
        Called when process completes, successful or not, before sig_finished is emitted.

        May do some or all of the followings:

         - Recombine data items split for parallel processing
         - Add output data/ROIs/extras to the IVM
         - log output from the worker using the ``log`` method
         - Set data items to be returned by ``output_data_items()``
        
        This method should not generally signal other components - they should connect
        to sig_finished instead
        """
        pass
    
    def split_args(self, n_workers, args):
        """
        Split input arguments into chunks for running in parallel
        
        This would normally called by run() after setting up the arguments to pass to the 
        worker run function.

        Note that this can be overridden to customize splitting behaviour
        
        :param args: Sequence of arguments to the worker run function. All must be pickleable objects.
                     By default Numpy arrays will be split along SPLIT_AXIS and a chunk passed to each
                     worker.
        :param n_workers: Number of parallel worker processes to use
        """
        # First argument is worker ID, second is queue
        split_args = [list(range(n_workers)), [self._queue,] * n_workers]

        for arg in args:
            if isinstance(arg, (np.ndarray, np.generic)):
                split_args.append(np.array_split(arg, n_workers, SPLIT_AXIS))
            else:
                split_args.append([arg,] * n_workers)

        # Transpose list of lists so first element is all the arguments for process 0, etc
        return list(map(list, zip(*split_args)))

    def recombine_data(self, data_list):
        """
        Recombine a sequence of data items into a single data item

        This implementation assumes data contains Numpy arrays and returns
        a new Numpy array concatenated along SPLIT_AXIS. However this method
        could be overridden, especially if split_data has been overridden.
        """
        shape = None
        for data_item in data_list:
            if data_item is not None:
                shape = data_item.shape
        if shape is None:
            raise RuntimeError("No data to re-combine")
        else:
            self.debug("Recombining data with shape: %s", shape)
        empty = np.zeros(shape)
        real_data = []
        for data_item in data_list:
            if data_item is None:
                real_data.append(empty)
            else:
                real_data.append(data_item)
        
        return np.concatenate(real_data, SPLIT_AXIS)

    def save_output(self, save_folder):
        """
        Save process output to a folder

        In practice this is very process dependent and this method may well need to be
        overridden. The default implementation uses the ``output_data_items()`` to
        get the names of the data items the process has created and writes these
        plus the logfile to the output folder
        """
        data_to_save = self.output_data_items()
        self.debug("Data to save: %s", data_to_save)    
        for d in data_to_save:
            qpdata = self.ivm.data.get(d, None)
            if qpdata is not None:
                save(qpdata, os.path.join(save_folder, d + ".nii"))
        logfile = open(os.path.join(save_folder, "logfile"), "w")
        logfile.write(self._log)
        logfile.close()

    @QtCore.Slot()
    def _complete(self):
        """
        Process completed
        
        Call finished method and emit sig_progress if successful, and 
        emit finished signal regardless. 
        """
        self.debug("Process completing, status=%i", self.status)
        self._completed = True
        if self.status == self.SUCCEEDED:
            try:
                self.finished(self._worker_output)
                self.sig_progress.emit(1)
            except Exception as exc:
                self.status = self.FAILED
                self.exception = exc
            
        # Get rid of all references to multprocessing workers and their output
        # this is necessary to avoid memory and process leakage
        if self._pool is not None:
            self._pool.close()
        self._pool = None
        self._workers = []
        self._queue = None
        self._worker_output = []
        self.debug("Emitting sig_finished")
        self.sig_finished.emit(self.status, self._log, self.exception)
        self._completed = True

    def _restart_timer(self):
        self._timer = threading.Timer(1, self._timer_cb)
        self._timer.daemon = True
        self._timer.start()

    def _timer_cb(self):
        if self.status == Process.RUNNING:
            self.timeout(self._queue)
            self._restart_timer()

    def _worker_finished_cb(self, result):
        worker_id, success, output = result
        self.debug("Process worker finished: id=%i, status=%s", worker_id, str(success))

        if self.status in (Process.FAILED, Process.CANCELLED):
            # If one process has already failed or been cancelled, ignore results of others
            self.debug("Ignoring worker, process already failed or cancelled")
            return
        elif success:
            self._workers[worker_id] = None # FIXME why? Memory leak?
            if worker_id < len(self._worker_output):
                self._worker_output[worker_id] = output
                if None not in self._worker_output:
                    self.status = Process.SUCCEEDED
        else:
            # If one process fails, they all fail. Output is just the first exception to be caught
            # FIXME cancel other workers?
            # FIXME log capture is ugly - better to have 'sig_failed' callback
            self.status = Process.FAILED
            self.exception = output
            if hasattr(output, "log"):
                self.log(output.log[:MAX_LOG_SIZE])
                if len(output.log) > MAX_LOG_SIZE:
                    self.log("WARNING: Exception log was too large - truncated at %i chars" % MAX_LOG_SIZE)

        if self.status != Process.RUNNING:
            # Need to use invokeMethod here because the process callback is in a 
            # different thread and the IVM (called by _complete) is not threadsafe
            self.metaObject().invokeMethod(self, "_complete", QtCore.Qt.QueuedConnection)



"""
Quantiphyse - Widgets for data simulation

Copyright (c) 2013-2018 University of Oxford
"""
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
from quantiphyse.gui.options import OptionBox, NumericOption, DataOption, OutputNameOption
from quantiphyse.gui.widgets import QpWidget, TitleWidget

from .processes import AddNoiseProcess, SimMotionProcess
    
class AddNoiseWidget(QpWidget):
    """
    Add noise to data
    """
    def __init__(self, **kwargs):
        super(AddNoiseWidget, self).__init__(name="Add noise", icon="noise", 
                                             desc="Add random noise to a data set", 
                                             group="Simulation", **kwargs)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Add Noise", help="noise")
        vbox.addWidget(title)

        self.option_box = OptionBox("Options")
        data = self.option_box.add("Data set", DataOption(self.ivm), key="data")
        self.option_box.add("Gaussian standard deviation", NumericOption(minval=0, maxval=100, default=50), key="std")
        self.option_box.add("Output name", OutputNameOption(src_data=data, suffix="_noisy"), key="output-name")
        vbox.addWidget(self.option_box)

        run_btn = QtGui.QPushButton('Run', self)
        run_btn.clicked.connect(self.run)
        vbox.addWidget(run_btn)
        
        vbox.addStretch(1)
 
    def batch_options(self):
        return "AddNoise", self.option_box.values()
        
    def run(self):
        options = self.batch_options()[1]
        process = AddNoiseProcess(self.ivm)
        process.execute(options)

class SimMotionWidget(QpWidget):
    """
    Widget to simulate random motion on a 4D data set
    """
    def __init__(self, **kwargs):
        super(SimMotionWidget, self).__init__(name="Simulate motion", icon="reg", 
                                              desc="Simulate random motion on a 4D data set", 
                                              group="Simulation", **kwargs)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Simulate Motion", help="sim_motion")
        vbox.addWidget(title)

        self.option_box = OptionBox("Options")
        data = self.option_box.add("Data set", DataOption(self.ivm, include_4d=True, include_3d=False), key="data")
        self.option_box.add("Motion standard deviation (mm)", NumericOption(minval=0, maxval=5, default=1, decimals=2), key="std")
        self.option_box.add("Padding (mm)", NumericOption(minval=0, maxval=10, default=5, decimals=1), key="padding", checked=True)
        self.option_box.add("Output name", OutputNameOption(src_data=data, suffix="_moving"), key="output-name")
        vbox.addWidget(self.option_box)

        run_btn = QtGui.QPushButton('Run', self)
        run_btn.clicked.connect(self.run)
        vbox.addWidget(run_btn)
        
        vbox.addStretch(1)

    def batch_options(self):
        return "SimMotion", self.option_box.values()
        
    def run(self):
        options = self.batch_options()[1]
        process = SimMotionProcess(self.ivm)
        process.execute(options)



"""
Quantiphyse - Functions for loading and querying plugins

Copyright (c) 2013-2018 University of Oxford
"""

import sys
import os
import glob
import importlib
import logging
import traceback

from quantiphyse.utils.local import get_local_file

PLUGIN_MANIFEST = None
LOG = logging.getLogger(__name__)

def _possible_module(mod_file):
    if os.path.basename(mod_file).startswith("_"):
        return None
    elif os.path.isdir(mod_file): 
        return os.path.basename(mod_file)
    elif mod_file.endswith(".py") or mod_file.endswith(".dll") or mod_file.endswith(".so"):
        return os.path.basename(mod_file).rsplit(".", 1)[0]

def _load_plugins_from_dir(dirname, pkgname, manifest):
    """
    Beginning of plugin system - load modules dynamically from the specified directory

    Then check in module for widgets and/or processes to return
    """
    LOG.debug("Loading plugins from %s", dirname)
    submodules = glob.glob(os.path.join(os.path.abspath(dirname), "*"))
    done = set()
    pythonpath = list(sys.path)
    try:
        sys.path.insert(0, dirname)
        for mod_file in submodules:
            mod = _possible_module(mod_file)
            if mod is not None and mod not in done:
                done.add(mod)
                try:
                    LOG.debug("Trying to import %s", mod)
                    module = importlib.import_module(mod, pkgname)
                    LOG.debug("Got %s (%s)", module.__name__, module.__file__)
                    if hasattr(module, "QP_WIDGETS"):
                        LOG.debug("Widgets found: %s %s", mod, module.QP_WIDGETS)
                        manifest["widgets"] = manifest.get("widgets", []) + module.QP_WIDGETS
                    if hasattr(module, "QP_PROCESSES"):
                        LOG.debug("Processes found: %s %s", mod, module.QP_PROCESSES)
                        manifest["processes"] = manifest.get("processes", []) + module.QP_PROCESSES
                    if hasattr(module, "QP_MANIFEST"):
                        # Module directories are added to the global PYTHONPATH
                        for deps_dir in module.QP_MANIFEST.pop("module-dirs", []):
                            deps_path = os.path.join(dirname, mod_file, deps_dir)
                            if os.path.isdir(deps_path):
                                pythonpath.append(deps_path)
                        # Everything else is added to the global manifest
                        for key, val in module.QP_MANIFEST.items():
                            LOG.debug("%s found: %s %s", key, mod, val)
                            manifest[key] = manifest.get(key, []) + val
                except ImportError:
                    LOG.warn("Error loading plugin: %s", mod)
                    traceback.print_exc()
    finally:
        sys.path = pythonpath

def _load_plugins_from_entry_points(manifest, key="quantiphyse_plugins"):
    import pkg_resources
    for ep in pkg_resources.iter_entry_points(key):
        for key, val in ep.load().items():
            LOG.debug("entry points: found: %s %s", key, val)
            manifest[key] = manifest.get(key, []) + val

def get_plugins(key=None, class_name=None):
    """
    Beginning of plugin system - load widgets dynamically from specified plugins directory
    """
    global PLUGIN_MANIFEST
    if PLUGIN_MANIFEST is None:
        PLUGIN_MANIFEST = {}

        plugin_dirs = {
            "quantiphyse.packages.core" : get_local_file("packages/core"), 
            "quantiphyse.packages.plugins" : get_local_file("packages/plugins"),
        }
        for pkg, plugin_dir in plugin_dirs.items():
            #if os.path.exists(plugin_dir):
            #    __import__(pkg)
            _load_plugins_from_dir(plugin_dir, pkg, PLUGIN_MANIFEST)
        _load_plugins_from_entry_points(PLUGIN_MANIFEST)
    
    if key is not None:
        plugins = PLUGIN_MANIFEST.get(key, [])
        if class_name is not None: 
            plugins = [p for p in plugins if p.__name__ == class_name]
    else:
        plugins = PLUGIN_MANIFEST
    return plugins
